from datetime import datetime, date, timedelta
from decimal import Decimal
from objict import objict
from io import StringIO
import importlib

from django.db import models as dm
from django.db.transaction import atomic
from django.apps import apps

from rest import helpers as rh
from rest import errors as re
from rest.encryption import ENCRYPTER, DECRYPTER
from rest import crypto
from rest import settings
from rest import search
from .metadata import MetaDataBase, MetaDataModel

DB_ROUTING_MAPS = settings.get("DB_ROUTING_MAPS", {})
TWO_DECIMAL_PLACES = Decimal(10) ** -2
EXCEPTION_ON_LIST_PERM_DENIED = settings.get("EXCEPTION_ON_LIST_PERM_DENIED", True)
ALLOW_BATCHING = settings.get("ALLOW_BATCHING", False)

GRAPH_HELPERS = objict()
GRAPH_HELPERS.restGet = None
GRAPH_HELPERS.get_request = None
GRAPH_HELPERS.views = None


class RestModel(object):
    class __RestMeta__:
        NO_SAVE_FIELDS = ["id", "pk", "created", "modified"]
        NO_SHOW_FIELDS = ["password"]
        WHITELISTED = ["merchant", "group", "user", "member"]

    class RestMeta:
        NO_SAVE_FIELDS = []
        SAVE_FIELDS = []
        GRAPHS = {}

    @staticmethod
    def generateUUID(*args, **kwargs):
        return crypto.generateUUID(*args, **kwargs)

    @classmethod
    def buildGraph(cls, name):
        # we need to build it
        if hasattr(cls.RestMeta, "GRAPHS"):
            graphs = cls.RestMeta.GRAPHS
            if isinstance(graphs, tuple):
                # this most likely means a typo in the definition, with an extra ","
                rh.log_error(f"{cls.__name__}.buildGraph({name}) is a tuple, check for comma after RestMeta.GRAPHS")
                graphs = graphs[0]
            if name not in graphs and name != "basic":
                name = "default"
            if name in graphs:
                graph = graphs[name]
            else:
                graph = {}
        else:
            graph = {}

        if "no_uscore" not in graph:
            graph["no_uscore"] = False

        no_show_fields = RestModel.__RestMeta__.NO_SHOW_FIELDS
        if hasattr(cls.RestMeta, "NO_SHOW_FIELDS"):
            no_show_fields = cls.RestMeta.NO_SHOW_FIELDS

        field_names = []
        for f in cls._meta.fields:
            if not f.name.endswith("_ptr"):
                if f.name not in no_show_fields:
                    field_names.append(f.name)

        if "graphs" in graph and name != "basic":
            if "recurse_into" not in graph:
                graph["recurse_into"] = []
            if "fields" in graph:
                graph["fields"] = graph["fields"]
            elif "fields" not in graph and "self" in graph["graphs"]:
                graph["fields"] = []
            else:
                graph["fields"] = field_names

            for field in graph["graphs"]:
                gname = graph["graphs"][field]
                size = None
                ForeignModel = None
                sort = None

                if field.startswith("generic__"):
                    if field not in graph["recurse_into"]:
                        graph["recurse_into"].append((field, gname))
                        continue

                if isinstance(gname, dict):
                    fm_name = gname.get("model")
                    gname = gname.get("graph")
                    if not gname:
                        gname = "default"
                    if fm_name:
                        a_name, m_name = fm_name.split(".")
                        ForeignModel = RestModel.getModel(a_name, m_name)

                if not field or field == "self":
                    # this means it is referencing self
                    foreign_graph = cls.buildGraph(gname)
                    for part in foreign_graph:
                        if part not in graph:
                            graph[part] = foreign_graph[part]
                        else:
                            for f in foreign_graph[part]:
                                if f not in graph[part]:
                                    graph[part].append(f)
                    continue

                if not ForeignModel:
                    ForeignModel = cls.get_fk_model(field)
                if not ForeignModel:
                    rh.log_print("no foreignkey: {0}".format(field))
                    continue

                if field not in graph["recurse_into"]:
                    graph["recurse_into"].append(field)

                if not hasattr(ForeignModel, "getGraph"):
                    # print "NO getGraph"
                    continue
                foreign_graph = ForeignModel.getGraph(gname)

                for part in ["fields", "recurse_into", "extra", "exclude"]:
                    if part not in foreign_graph:
                        continue
                    graph_part = foreign_graph[part]
                    if part not in graph:
                        graph[part] = []
                    root_part = graph[part]
                    for f in graph_part:
                        if type(f) is tuple:
                            f1, f2 = f
                            nfname = ("{0}.{1}".format(field, f1), f2)
                        elif graph["no_uscore"] and '_' in f:
                            f1, f2 = f, f.replace('_', '').split('.')[-1]
                            nfname = ("{0}.{1}".format(field, f1), f2)
                        else:
                            nfname = "{0}.{1}".format(field, f)
                        if nfname not in root_part:
                            root_part.append(nfname)
            del graph["graphs"]
        elif "graphs" in graph:
            del graph["graphs"]

        if "fields" not in graph:
            if graph["no_uscore"]:
                graph["fields"] = []
                for f in field_names:
                    if "_" in f:
                        f1, f2 = f, f.lower().replace('_', '')
                        graph["fields"].append((f1, f2))
                    else:
                        graph["fields"].append(f)
            else:
                graph["fields"] = field_names

        if "no_uscore" in graph:
            del graph["no_uscore"]
        return graph

    @classmethod
    def ro_objects(cls):
        using = getattr(cls.RestMeta, "RO_DATABASE", None)
        if using is None:
            using = getattr(cls.RestMeta, "DATABASE", None)
        # if using is None:
        #   if settings.DATABASES.get("readonly", None) != None:
        #       using = "readonly"
        if using:
            using = cls.get_db_mapping(using)
            return cls.objects.using(using)
        return cls.objects

    @classmethod
    def rw_objects(cls):
        using = getattr(cls.RestMeta, "DATABASE", None)
        if using:
            using = cls.get_db_mapping(using)
            return cls.objects.using(using)
        return cls.objects

    @atomic
    def saveNow(self, **kwargs):
        return self.save(**kwargs)

    def safeSave(self, **kwargs):
        using = getattr(self.RestMeta, "DATABASE", None)
        if using:
            using = self.get_db_mapping(using)
            return self.save(using=using, **kwargs)
        return self.save(**kwargs)

    @classmethod
    def getGraph(cls, name):
        if name is None:
            name = "default"
        graph_key = f"_graph_{name}__"
        if hasattr(cls._meta, graph_key):
            return getattr(cls._meta, graph_key)
        graph = cls.buildGraph(name)
        setattr(cls._meta, graph_key, graph)
        return graph

    def toGraph(self, request=None, graph="basic"):
        RestModel._setupGraphHelpers()
        if not request:
            request = GRAPH_HELPERS.get_request()
        return GRAPH_HELPERS.restGet(request, self, return_httpresponse=False, **self.getGraph(graph))

    def getFieldValue(self, name, default=None):
        fields = name.split('.')
        obj = self
        for f in fields:
            if not hasattr(obj, f):
                return default
            obj = getattr(obj, f, None)
            if obj is None:
                return obj
        return obj

    @classmethod
    def getActiveLogger(cls):
        return rh.getLogger(cls.getActiveRequest())

    @classmethod
    def getActiveMember(cls):
        request = cls.getActiveRequest()
        if request:
            return request.member
        return None

    @classmethod
    def getActiveRequest(cls):
        return rh.getActiveRequest()

    @classmethod
    def getFromRequest(cls, request):
        key = cls.get_class_name().lower()
        key_p = "{0}_id".format(key)
        lookup_fields = [key, key_p]
        using = getattr(cls.RestMeta, "DATABASE", None)
        for field in lookup_fields:
            value = request.DATA.get(field)
            if value:
                if not using:
                    obj = cls.objects.filter(pk=value).first()
                else:
                    obj = cls.objects.using(using).filter(pk=value).first()
                if obj:
                    return obj
        lookup_fields = getattr(cls.RestMeta, "UNIQUE_LOOKUP", [])
        for field in lookup_fields:
            value = request.DATA.get([field, "{}_{}".format(key, field)])
            if value:
                q = {}
                q[field] = value
                if not using:
                    obj = cls.objects.filter(**q).first()
                else:
                    obj = cls.objects.using(using).filter(**q).first()
                if obj:
                    return obj
        return None

    @classmethod
    def getFromPK(cls, pk):
        if isinstance(pk, dict):
            fields = pk
            if "id" in fields:
                pk = fields.get("id")
            elif "pk" in fields:
                pk = fields.get("pk")
            else:
                lookup_fields = getattr(cls.RestMeta, "UNIQUE_LOOKUP", [])
                for field in lookup_fields:
                    if field in fields:
                        pk = fields.get(field)
                        q = {}
                        q[field] = pk
                        return cls.objects.filter(**q).first()
            pk = None
        if pk in (None, ""):
            return None
        using = getattr(cls.RestMeta, "DATABASE", None)
        if using:
            return cls.objects.using(using).filter(pk=pk).first()
        if isinstance(pk, str) and pk.isdigit():
            pk = int(pk)

        if isinstance(pk, int):
            return cls.objects.filter(pk=pk).first()
        # concerns with this being misused
        lookup_fields = getattr(cls.RestMeta, "UNIQUE_LOOKUP", [])
        for field in lookup_fields:
            q = {}
            q[field] = pk
            obj = cls.objects.filter(**q).first()
            if obj is not None:
                return obj
        return None

    @classmethod
    def restEncrypt(cls, data):
        if ENCRYPTER:
            return ENCRYPTER.encrypt(data)
        return data

    @staticmethod
    def restGetModel(app_name, model_name):
        return apps.get_model(app_name, model_name)

    @staticmethod
    def getModel(app_name, model_name):
        return apps.get_model(app_name, model_name)

    @staticmethod
    def getModelInstance(app_name, model_name, **kwargs):
        Model = apps.get_model(app_name, model_name)
        return Model.objects.filter(**kwargs).last()

    @staticmethod
    def createModelInstance(app_name, model_name, **kwargs):
        Model = apps.get_model(app_name, model_name)
        return Model(**kwargs)

    def restGetGenericModel(self, field):
        # called by the rest module to magically parse
        # a component that is marked genericr elation in a graph
        if not hasattr(self, field):
            rh.log_print("model has no field: {0}".format(field))
            return None

        name = getattr(self, field)
        if not name or "." not in name:
            return None
        a_name, m_name = name.split(".")
        model = RestModel.getModel(a_name, m_name)
        if not model:
            rh.log_print("GENERIC MODEL DOES NOT EXIST: {0}".format(name))
        return model

    def restGetGenericRelation(self, field):
        # called by the rest module to magically parse
        # a component that is marked genericrelation in a graph
        GenericModel = self.restGetGenericModel(field)
        if not GenericModel:
            return None
        # verify user has permission to access this model
        key = getattr(self, "{0}_id".format(field))
        obj = GenericModel.rw_objects().filter(pk=key).first()
        if obj is None:
            return None
        if hasattr(obj, "on_rest_can_get"):
            if not obj.on_rest_can_get(self.getActiveRequest()):
                return None
        return obj

    @classmethod
    def get_db_mapping(cls, name):
        if name in DB_ROUTING_MAPS:
            return DB_ROUTING_MAPS[name]
        return name

    @classmethod
    def restGetModelDB(cls, default=None):
        if hasattr(cls, "RestMeta"):
            return cls.get_db_mapping(getattr(cls.RestMeta, "DATABASE", default))
        return default

    @property
    def has_model_changed(self):
        if hasattr(self, "_changed__"):
            return len(self._changed__) > 0
        return False

    def saveFields(self, allow_null=True, **kwargs):
        """
        Helper method to save a list of fields
        """
        self._changed__ = objict()
        for key, value in list(kwargs.items()):
            if value is None and not allow_null:
                continue
            self.restSaveField(key, value)
        if len(self._changed__):
            self.save()

    def restSaveField(self, fieldname, value, has_fields=False, has_no_fields=False, using=None):
        if not hasattr(self, "_changed__"):
            self._changed__ = objict()

        if fieldname.startswith("_"):
            return
        if not hasattr(self, "_field_names__"):
            self._field_names__ = [f.name for f in self._meta.get_fields()]
        # if not hasattr(self, "_related_field_names__"):
        #     self._related_field_names__ = self.get_related_name_fields()
        # print "saving field: {0} = {1}".format(fieldname, value)
        if fieldname in RestModel.__RestMeta__.NO_SAVE_FIELDS:
            return
        if has_no_fields and fieldname in self.RestMeta.NO_SAVE_FIELDS:
            return
        if has_fields and fieldname not in self.RestMeta.SAVE_FIELDS:
            return
        if fieldname.endswith("_id") and not self.get_field_type(fieldname):
            # django will have ForeignKeys with _id, we don't want that, on_delete=dm.CASCADE
            fieldname = fieldname[:-3]
        setter = F"set_{fieldname}"
        if hasattr(self, setter):
            getattr(self, setter)(value)
            return

        if fieldname in self._field_names__:
            # TODO check if it is a function
            if isinstance(value, dm.Model):
                setattr(self, fieldname, value)
                self._changed__[fieldname] = True
                return
            ForeignModel = self.get_fk_model(fieldname)
            if ForeignModel:
                if isinstance(value, dict):
                    obj = getattr(self, fieldname, None)
                    if obj is None:
                        if hasattr(ForeignModel, "getFromPK"):
                            obj = ForeignModel.getFromPK(value)
                        if obj is None:
                            obj = ForeignModel()
                    else:
                        if hasattr(ForeignModel, "getFromPK"):
                            new_obj = ForeignModel.getFromPK(value)
                        if new_obj is None:
                            lookup_fields = getattr(ForeignModel.RestMeta, "UNIQUE_LOOKUP", [])
                            if "id" in value or any(name in value for name in lookup_fields):
                                obj = ForeignModel()
                        elif new_obj.pk != obj.pk:
                            obj = new_obj
                    if using is None:
                        using = self.restGetModelDB()
                    obj.checkPermsAndSave(None, value, using=using)
                    # rh.log_print("{} vs {}".format(self._state.db, obj._state.db))
                    # rh.log_print("saving FK to {} ({}.{}) - {}".format(fieldname, using, obj.pk, type(obj)), value)
                    setattr(self, fieldname, obj)
                    self._changed__[fieldname] = True
                    return
                elif hasattr(ForeignModel, "getFromPK"):
                    value = ForeignModel.getFromPK(value)
                elif isinstance(value, int) or (isinstance(value, str) and value.isdigit()):
                    value = ForeignModel.objects.filter(pk=int(value)).first()
                elif "MediaItem" in ForeignModel.__name__:
                    if value:
                        self.saveMediaFile(value, fieldname, None, True)
                    return
                elif not bool(value):
                    value = None
            else:
                # field_model, model, direct, mm = self._meta.get_field_by_name(fieldname)
                field_model = self._meta.get_field(fieldname)
                # hack to handle save datetime fields correctly from floats
                try:
                    if field_model and value != None:
                        field_model_name = field_model.__class__.__name__
                        if field_model_name == "DateTimeField":
                            value = rh.parseDateTime(value)
                            # value = datetime.fromtimestamp(float(value))
                        elif field_model_name == "DateField":
                            value = rh.parseDate(value, as_date=True)
                        elif field_model_name == "IntegerField":
                            value = int(value)
                        elif field_model_name == "FloatField":
                            value = float(value)
                        elif field_model_name == "CurrencyField":
                            value = Decimal(value).quantize(TWO_DECIMAL_PLACES)
                        elif field_model_name == "BooleanField":
                            if value in [True, 1, 'True', 'true', '1', 't', 'y', 'yes']:
                                value = True
                            else:
                                value = False
                except Exception:
                    return
            if hasattr(self, fieldname) and getattr(self, fieldname) != value:
                self._changed__[fieldname] = getattr(self, fieldname)
            setattr(self, fieldname, value)

    def saveFromRequest(self, request, **kwargs):
        if "files" not in kwargs:
            kwargs["files"] = request.FILES
        return self.checkPermsAndSave(request, request.DATA, **kwargs)

    def hasFieldChanged(self, fieldname):
        if not hasattr(self, "_changed__"):
            return False
        return fieldname in self._changed__

    def _recordRestChange(self, fieldname, old_value):
        if not hasattr(self, "_changed__"):
            self._changed__ = objict()
        if "." in fieldname:
            fields = fieldname.split('.')
            root = self._changed__
            for f in fields[:-1]:
                if f not in root:
                    root[f] = objict()
                root = root[f]
            root[fields[-1]] = old_value
        else:
            self._changed__[fieldname] = old_value

    def checkPermsAndSave(self, request, data, files=None, **kwargs):
        can_save = getattr(self.RestMeta, "CAN_SAVE", True)
        if not can_save:
            raise re.PermissionDeniedException("saving not allowed via rest for this model.")
        # check check for save permissions
        if request is None:
            request = RestModel.getActiveRequest()
            if request is None:
                request = objict(member=None, FILES=[], DATA=objict(), is_authenticated=False, is_local=True)
        self.onRestCanSave(request, data, kwargs)
        return self.saveFromDict(request, data, files=None, **kwargs)

    def saveFromDict(self, request, data, files=None, **kwargs):
        save_media_files = getattr(self.RestMeta, "SAVE_MEDIA_FILES", False) or getattr(self.RestMeta, "SAVE_MEDIA_FILES_BY_NAME", False)
        is_new = self.id is None
        has_fields = hasattr(self.RestMeta, "SAVE_FIELDS") and len(self.RestMeta.SAVE_FIELDS)
        has_no_fields = hasattr(self.RestMeta, "NO_SAVE_FIELDS") and len(self.RestMeta.NO_SAVE_FIELDS)
        self._field_names__ = [f.name for f in self._meta.get_fields()]
        # fix for multidatabase support and using readonly db for get
        self._state.db = kwargs.get("using", self.restGetModelDB("default"))
        auto_save_fields = getattr(self.RestMeta, "AUTO_SAVE", None)
        if auto_save_fields:
            # rh.log_print(auto_save_fields)
            for field in auto_save_fields:
                # rh.log_print(field)
                if isinstance(field, tuple):
                    m_field, req_field = field
                else:
                    m_field = field
                    req_field = field
                if request is not None:
                    req_value = getattr(request, req_field, None)
                    if req_value is not None:
                        data[m_field] = req_value
            # rh.log_print(data)
        self._changed__ = objict()
        if hasattr(self.RestMeta, "POST_SAVE_FIELDS"):
            post_save_fields = self.RestMeta.POST_SAVE_FIELDS
        else:
            post_save_fields = []
        using = kwargs.get("using", self.restGetModelDB())
        deferred = {}
        group_fields = {}
        for fieldname in data:
            # we allow override via kwargs
            value = data.get(fieldname)
            if "." in fieldname:
                gname = fieldname[:fieldname.find('.')]
                fname = fieldname[fieldname.find('.')+1:]
                setter = "set_{0}".format(gname)
                if hasattr(self, setter):
                    if gname not in group_fields:
                        group_fields[gname] = {}
                    group_fields[gname][fname] = value
                    continue
            if fieldname in post_save_fields or fieldname.startswith("metadata"):
                deferred[fieldname] = value
                continue
            if fieldname not in kwargs:
                self.restSaveField(fieldname, value, has_fields, has_no_fields, using=using)
        for key, value in list(kwargs.items()):
            if key in post_save_fields:
                deferred[fieldname] = value
                continue
            self.restSaveField(key, value, has_fields, has_no_fields, using=using)

        rfiles = self.restSaveFiles(request, files)
        if getattr(self.RestMeta, "ALWAYS_SAVE", True) or self.has_model_changed:
            self.on_rest_pre_save(request)
            self.save(using=using)
        for key, value in list(deferred.items()):
            self.restSaveField(key, value, has_fields, has_no_fields, using=using)

        if len(deferred):
            self.save(using=using)

        if rfiles is not None and len(rfiles) and save_media_files:
            self.restSaveMediaFiles(rfiles)

        # these setters are responsible for saving themselves
        for gname in group_fields:
            setter = "set_{0}".format(gname)
            getattr(self, setter)(request, group_fields[gname])

        self.on_rest_saved(request, is_new=is_new)
        if is_new:
            self.on_rest_created(request)
        return self

    def restSaveFiles(self, request, files=None):
        if files is None and request is not None:
            files = request.FILES
        if not files:
            return files
        remaining_files = {}
        for name in files:
            key = f"upload__{name}"
            if hasattr(self, key):
                getattr(self, key)(files[name], name)
            else:
                ForeignModel = self.get_fk_model(name)
                if ForeignModel and ForeignModel.__name__ == "MediaItem":
                    # rh.log_print("saving media file: {}".format(name))
                    self.saveMediaFile(files[name], name, files[name].name)
                else:
                    remaining_files[name] = files[name]
        return remaining_files

    def restSaveMediaFiles(self, files=None):
        MediaItemRef = RestModel.getModel("medialib", "MediaItemRef")
        component = self.get_class_name(True)
        # rh.log_print("saving media files refs: {}".format(component))
        for name in files:
            mi = self.saveMediaFile(files[name], name, is_local=False)
            mr = None
            if getattr(self.RestMeta, "SAVE_MEDIA_FILES_BY_NAME", False):
                # use existing reference
                mr = MediaItemRef.objects.filter(component=component, component_id=self.id, media__name=name).last()
                if mr:
                    rh.log_print(F"existing media files refs: {name}.{component}.{self.id}")
                    mr.media = mi
                    mr.save()
            if mr is None:
                mr = MediaItemRef(media=mi, component=component, component_id=self.id)
                mr.save()

    def getMediaFiles(self):
        MediaItemRef = RestModel.getModel("medialib", "MediaItemRef")
        component = self.get_class_name(True)
        return MediaItemRef.objects.filter(component=component, component_id=self.id)

    def media_files(self, graph="simple"):
        MediaItemRef = RestModel.getModel("medialib", "MediaItemRef")
        component = self.get_class_name(True)
        qset = MediaItemRef.objects.filter(component=component, component_id=self.id)
        if not getattr(self.RestMeta, "SAVE_MEDIA_FILES_BY_NAME", False):
            return MediaItemRef.restList(self.getActiveRequest(), qset, graph, None, False)

        out = {}
        for mr in qset:
            out[mr.media.name] = mr.media.toDict("basic")
        return out

    def changesFromDict(self, data):
        deltas = []
        field_names = [f.name for f in self._meta.get_fields()]
        for key in data:
            if key not in field_names:
                continue
            # we allow override via kwargs
            value = data.get(key)

            # field_model, model, direct, mm = self._meta.get_field_by_name(key)
            field_model = self._meta.get_field(key)
            # hack to handle save datetime fields correctly from floats
            try:
                if field_model and value != None:
                    field_model_name = field_model.__class__.__name__
                    if field_model_name == "DateTimeField":
                        value = datetime.fromtimestamp(float(value))
                    elif field_model_name == "DateField":
                        value = rh.parseDate(value)
                    elif field_model_name == "IntegerField":
                        value = int(value)
                    elif field_model_name == "FloatField":
                        value = float(value)
                    elif field_model_name == "CurrencyField":
                        value = Decimal(value).quantize(TWO_DECIMAL_PLACES)
                if hasattr(self, key) and getattr(self, key) != value:
                    deltas.append(key)
            except:
                pass
        return deltas

    def copyFieldsFrom(self, obj, fields):
        for f in fields:
            if hasattr(self, f):
                setattr(self, f, getattr(obj, f))

    def saveMediaFile(self, file, name, file_name=None, is_base64=False, group=None, is_local=True):
        """
        Generic method to save a media file
        """
        if file_name is None:
            file_name = name
        MediaItem = RestModel.getModel("medialib", "MediaItem")
        # make sure we set the name base64_data
        if is_base64:
            mi = MediaItem(name=file_name, base64_data=file, group=group)
        elif isinstance(file, str):
            if file.startswith("https:") or file.startswith("http:"):
                mi = MediaItem(name=file_name, downloadurl=file, group=group)
            else:
                file = StringIO(file)
                file.size = len(file)
                file.name = file_name
                mi = MediaItem(name=file_name, group=group, newfile=file)
        else:
            mi = MediaItem(name=file_name, newfile=file, group=group)
        # rh.log_print(F"saving media file: {name}")
        mi.save()
        if is_local:
            setattr(self, name, mi)
        return mi

    def updateLogModel(self, request, model):
        if not request:
            request = self.getActiveRequest()
        if not request or not hasattr(request, "setLogModel"):
            rh.log_print("request does not support setLogModel")
            return
        if not self.id:
            self.save()
        request.setLogModel(model, self.id)

    def checkIsOwner(self, member):
        owner_field = getattr(self.RestMeta, "OWNER_FIELD", None)
        if owner_field is None:
            owner_field = getattr(self.RestMeta, "VIEW_PERMS_MEMBER_FIELD", None)
        if owner_field is not None:
            owner = getattr(self, owner_field, None)
            return member == owner
        return False

    def on_rest_pre_get(self, request):
        pass

    def on_rest_can_get(self, request):
        if request is None:
            return True
        perms = getattr(self.RestMeta, "VIEW_PERMS", None)
        if perms:
            if "owner" in perms and self.checkIsOwner(request.member):
                return True
            # we need to check if this user has permission
            group_field = getattr(self.RestMeta, "GROUP_FIELD", "group")
            status, error, code = rh.requestHasPerms(request, perms, getattr(self, group_field, None))
            if not status:
                raise re.PermissionDeniedException(error + " getting", code)
        return True

    def on_rest_get(self, request):
        # check view permissions
        if not self.on_rest_can_get(request):
            return self.restStatus(request, False, error="permission denied", error_code=402)
        self.on_rest_pre_get(request)
        return self.restGet(request)

    def onRestCanDelete(self, request=None, data=None, extended=None):
        self.onRestCheckSavePerms(request, data, extended)

    def onRestCanSave(self, request=None, data=None, extended=None):
        self.onRestCheckSavePerms(request, data, extended)

    def onRestCheckSavePerms(self, request=None, data=None, extended=None):
        if request is None:
            request = self.getActiveRequest()
            if request is None:
                return True
        if getattr(request, "is_local", False):
            return True
        perms = getattr(self.RestMeta, "SAVE_PERMS", getattr(self.RestMeta, "VIEW_PERMS", None))
        if self.id is None:
            perms = getattr(self.RestMeta, "CREATE_PERMS", perms)
        if perms:
            if "owner" in perms and self.checkIsOwner(request.member):
                return True
            # we need to check if this user has permission
            group_field = getattr(self.RestMeta, "GROUP_FIELD", "group")
            # rh.log_error(F"group field={group_field}")
            if group_field == "self" and self.pk:
                group = self
            elif self.id is None:
                group = None
                if extended is not None:
                    group = extended.get(group_field, None)
                if group is None:
                    if data is not None:
                        group = data.get(group_field, None)
                if group is None:
                    if request is not None:
                        group = request.DATA.get(group_field, None)
            elif "__" in group_field and hasattr(self, "group"):
                group = self.group
            elif "__" in group_field:
                fields = group_field.split("__")
                group = self
                for field in fields:
                    group = getattr(group, field, None)
                    if group is None:
                        break
            else:
                group = getattr(self, group_field, None)
            status, error, code = rh.requestHasPerms(request, perms, group)
            if not status:
                if self.id is None and code == 402:
                    code = 435
                    rh.log_error(F"permission denied onRestCheckSavePerms: {self.__class__.__name__}\ngroup: {group}\nstatus={status}, error={error}, code={code}\nperms:", perms)
                raise re.PermissionDeniedException(error + " saving", code)

    def on_rest_post(self, request):
        self.checkPermsAndSave(request, request.DATA)
        status_only = request.DATA.get("status_only", False, field_type=bool)
        if status_only:
            return self.restStatus(request, True)
        graph = request.DATA.get("graph", "default")
        return self.restGet(request, graph)

    def on_rest_pre_save(self, request, **kwargs):
        pass

    def on_rest_created(self, request):
        pass

    def on_rest_saved(self, request, is_new=False):
        pass

    def on_rest_delete(self, request, force=False):
        can_delete = getattr(self.RestMeta, "CAN_DELETE", False)
        if not can_delete:
            raise re.PermissionDeniedException(f"deletion not allowed for {self.get_class_name()}", 438)
        if not force:
            self.onRestCanDelete(request)
        self.on_rest_deleted(request)
        self.delete()
        RestModel._setupGraphHelpers()
        return self.restStatus(request, True)

    def on_rest_deleted(self, request):
        self.auditLog(F"deleted {self.pk}", "deleted", level=8)
        if request and request.member:
            request.member.auditLog(F"deleted {self.get_class_name(True)}:{self.pk}", "deleted", level=8)

    def auditLog(self, message, action="log", path=None, level=0, group=None, method=None):
        if group is None and hasattr(self, "group"):
            group = self.group
        PLOG = self.getModel("auditlog", "PersistentLog")
        component = self.get_class_name(True)
        PLOG.log(message=message, action=action, path=path, level=level, method=method, component=component, pkey=self.id, group=group)

    @classmethod
    def restList(cls, request, qset, graph=None, totals=None, return_httpresponse=True):
        RestModel._setupGraphHelpers()
        sort = request.DATA.get("sort", getattr(cls.RestMeta, "DEFAULT_SORT", "-id"))
        if sort:
            # make sure we have the sort field
            tsort = sort
            if tsort.startswith("-"):
                tsort = sort[1:]
            if "__" not in tsort and not cls.hasField(tsort):
                sort = "-id"

        if totals:
            fields = totals
            totals = {}
            for tf in fields:
                cls_method = "qset_totals_{}".format(tf)
                if hasattr(cls, cls_method):
                    totals[tf] = getattr(cls, cls_method)(qset, request)
        if not graph and request is not None:
            graph = request.DATA.get("graph", "default")
        return GRAPH_HELPERS.restList(
            request, qset, sort=sort, totals=totals, 
            return_httpresponse=return_httpresponse,
            response_params=dict(graph=graph),
            **cls.getGraph(graph))

    @classmethod
    def restListSort(cls, request, qset):
        from rest.serializers.collection import sort_list
        sort = request.DATA.get("sort", getattr(cls.RestMeta, "DEFAULT_SORT", "-id"))
        sort = request.DATA.get("sort")
        if sort:
            qset, sort_args = sort_list(qset, sort)
        return qset

    @classmethod
    def toList(cls, qset, graph=None, totals=None, request=None):
        if request is None:
            request = cls.getActiveRequest()
        return cls.restList(request, qset, graph, totals, False)

    def restStatus(self, request, status, **kwargs):
        RestModel._setupGraphHelpers()
        return GRAPH_HELPERS.restStatus(request, status, **kwargs)

    def restGet(self, request, graph=None, as_dict=False):
        RestModel._setupGraphHelpers()
        if not request:
            request = self.getActiveRequest()
        if not graph and request:
            graph = request.DATA.get("graph", "default")
        elif not graph:
            graph = "default"
        return_response = not as_dict
        return GRAPH_HELPERS.restGet(request, self, return_httpresponse=return_response, **self.getGraph(graph))

    def toDict(self, graph=None, schema=None):
        RestModel._setupGraphHelpers()
        if schema is None:
            return self.restGet(None, graph=graph, as_dict=True)
        return GRAPH_HELPERS.restGet(None, self, return_httpresponse=False, **schema)

    def __str__(self):
        return f"<{self.get_class_name(True)}>"

    @classmethod
    def on_rest_list_filter(cls, request, qset=None):
        # override on do any pre filters
        return cls.on_rest_list_perms(request, qset)

    @classmethod
    def on_rest_filter_member_field(cls, request, qset):
        member_field = getattr(cls.RestMeta, "VIEW_PERMS_MEMBER_FIELD", None)
        if member_field is None:
            # this would be permission denied
            member_field = getattr(cls.RestMeta, "OWNER_FIELD", None)
        if member_field is None:
            if EXCEPTION_ON_LIST_PERM_DENIED:
                raise re.PermissionDeniedException(f"user does not have permission to access {cls.get_class_name()}")
            return cls.objects.none()
        # rh.log_error(dict(**{member_field: request.member}))
        return qset.filter(**{member_field: request.member})

    @classmethod
    def on_rest_list_has_perms(cls, request, perms, qset=None):
        if request.auth_model is not None:
            if request.auth_model.hasPerm(perms):
                return True
        if request.member is None:
            return False
        if request.member.hasPerm(perms):
            return True
        if request.group and request.member.hasGroupPerm(request.group, perms):
            return True
        return False

    @classmethod
    def on_rest_list_perms(cls, request, qset=None):
        perms = getattr(cls.RestMeta, "LIST_PERMS_GROUP", None)
        if perms is None:
            perms = getattr(cls.RestMeta, "VIEW_PERMS", None)
        has_perms = True
        if perms is not None:
            has_perms = cls.on_rest_list_has_perms(request, perms, qset)
        
        if request.group:
            qset = cls.on_rest_filter_children(request, qset)
        if not has_perms:
            return cls.on_rest_filter_member_field(request, qset)
        return qset

    @classmethod
    def on_rest_filter_children(cls, request, qset=None):
        # rh.log_error("filter by group?")
        group_field = getattr(cls.RestMeta, "GROUP_FIELD", "group")
        if group_field is None or ("__" not in group_field and not cls.has_model_field_name(group_field)):
            return qset
        parent_kinds = getattr(cls.RestMeta, "LIST_PARENT_KINDS", ["org"])
        if request.DATA.get("child_groups") or request.group.kind in parent_kinds:
            list_depth = getattr(cls.RestMeta, "LIST_CHILD_DEPTH", 0)
            ids = request.group.getAllChildrenIds(depth=list_depth)
            ids.append(request.group.id)
            # to avoid future filtering issues remove group
            request.group = None
            request.DATA.remove(group_field)
            if group_field != "group":
                # rh.debug("removing group field!!")
                request.DATA.remove("group")
            q = {}
            q["{}_id__in".format(group_field)] = ids
            return qset.filter(**q)
        elif "__" in group_field:
            q = {}
            q[group_field] = request.group.id
            qset = qset.filter(**q)
            # rh.log_error(q, qset.count())
        return qset

    @classmethod
    def on_rest_list_ready(cls, request, qset=None):
        # override on do any post filters
        # rh.log_error("RAW SQL")
        # rh.log_error(qset.query)
        return qset

    @classmethod
    def on_rest_date_filter(cls, request, qset=None):
        dr_start = request.DATA.get("dr_start", field_type=datetime)
        dr_end = request.DATA.get("dr_end", field_type=datetime)
        dr_end_str = request.DATA.get("dr_end")
        tr_start = request.DATA.get("tr_start")
        tr_end = request.DATA.get("tr_end")
        dr_offset = request.DATA.get("dr_offset", 0, field_type=int)
        dr_tz = request.DATA.get("dr_timezone")

        # if we only specify timerange lets set the start to today
        if dr_start is None:
            if tr_start is not None:
                dr_start = datetime.now()

        if dr_start is None:
            return qset

        if dr_end is None:
            dr_end = datetime.now() + timedelta(hours=5)
        elif dr_end == dr_start and not tr_end:
            dr_end = dr_start + timedelta(days=1)
        elif "-" in dr_end_str or "/" in dr_end_str:
            dr_end = dr_end + timedelta(days=1)

        if tr_start is not None:
            dr_start = rh.updateTimeFromString(dr_start, tr_start)
        if tr_end is not None:
            dr_end = request.DATA.get("dr_end", field_type=datetime)
            dr_end = rh.updateTimeFromString(dr_end, tr_end)

        if dr_tz is not None:
            dr_start = request.DATA.get("dr_start", field_type=datetime)
            dr_end = request.DATA.get("dr_end", field_type=datetime)
            dr_eod = request.DATA.get("dr_eod", 0, field_type=int)
            dr_granularity = request.DATA.get("granularity", "day")
            dr_start, dr_end = rh.getDateRange(dr_start, dr_end, dr_granularity, dr_tz, hour=dr_eod)
        elif dr_offset > 0:
            dr_start = dr_start + timedelta(minutes=dr_offset)
            dr_end = dr_end + timedelta(minutes=dr_offset)
    
        rh.debug("tr_end changing", str(dr_start), str(dr_end))
        dr_field = request.DATA.get("dr_field", getattr(cls.RestMeta, "DATE_RANGE_FIELD", "created"))
        q = dict()
        q["{}__gte".format(dr_field)] = dr_start
        q["{}__lte".format(dr_field)] = dr_end
        qset = qset.filter(**q)
        return qset

    @classmethod
    def on_rest_list(cls, request, qset=None):
        qset = cls.on_rest_list_query(request, qset)
        graph = request.DATA.get("graph", "list")
        format = request.DATA.get("format")
        if format:
            return cls.on_rest_list_format(request, format, qset)
        totals = request.DATA.getlist("totals", None)
        return cls.restList(request, qset, graph, totals)

    @classmethod
    def on_rest_list_query(cls, request, qset=None):
        cls._boundRest()
        cls.on_rest_set_query_defaults(request)
        request.rest_class = cls
        if qset is None:
            qset = cls.ro_objects().all()
        qset = cls.on_rest_list_filter(request, qset)
        qset = cls.filterFromRequest(request, qset)
        qset = cls.queryFromRequest(request, qset)
        qset = cls.searchFromRequest(request, qset)
        qset = cls.on_rest_date_filter(request, qset)
        qset = cls.on_rest_list_ready(request, qset)
        if request.default_rest_filters:
            qset = qset.filter(**request.default_rest_filters)
        return qset

    @classmethod
    def getSortArgs(cls, sort):
        if not bool(sort):
            return []

        elif sort.endswith("_display"):
            # fix for django _display kinds being sorted
            sort = sort[:sort.find("_display")]
        sort_args = []
        for s in sort.split(","):
            if "metadata" in s:
                continue
            if s.endswith("_display"):
                # s = s[:s.find("_display")]
                continue
            s = s.replace('.', '__')
            sort_args.append(s)
        return sort_args

    @classmethod
    def getRestFormatFields(cls, format, graph=None):
        fields = []
        if hasattr(cls.RestMeta, "FORMATS"):
            fields = cls.RestMeta.FORMATS.get(format, fields)
        if len(fields) == 0:
            if format == "json":
                g = cls.getGraph("default")
                if "fields" in g:
                    return g["fields"]
            no_show_fields = RestModel.__RestMeta__.NO_SHOW_FIELDS
            if hasattr(cls.RestMeta, "NO_SHOW_FIELDS"):
                no_show_fields = cls.RestMeta.NO_SHOW_FIELDS
            for f in cls._meta.fields:
                if not f.name.endswith("_ptr"):
                    if f.name not in no_show_fields:
                        fields.append(f.name)
        return fields

    @classmethod
    def on_rest_list_format(cls, request, format, qset):
        if format in ["summary", "summary_only"]:
            return cls.on_rest_list_summary(request, qset)
        fields = None
        if format == "json":
            g = cls.getGraph(request.DATA.get("graph", "download"))
            if g and "fields" in g:
                fields = g["fields"]
        if fields is None:
            fields = cls.getRestFormatFields(format)
        if fields or format == "json":
            name = request.DATA.get("format_filename", None)
            format_size = request.DATA.get("format_size", 10000)
            localize = request.DATA.get("localize", None, field_type=dict)
            if name is None:
                ext = format
                if "_" in ext:
                    ext = ext[:ext.find("_")]
                name = "{}.{}".format(cls.get_class_name(), ext)
            # print "csv size: {}".format(qset.count())
            sort = request.DATA.get("sort", getattr(cls.RestMeta, "DEFAULT_SORT", "-id"))
            sort_args = cls.getSortArgs(sort)
            if sort_args:
                try:
                    qset = qset.order_by(*sort_args)
                except Exception:
                    pass
            cls._boundRest()
            if format.startswith("json"):
                return GRAPH_HELPERS.views.restJSON(
                    request, qset, fields, 
                    name, format_size, localize=localize)
            elif format.startswith("csv"):
                return GRAPH_HELPERS.views.restCSV(
                    request, qset, fields,
                    name, format_size, localize=localize)
            elif format == "xlsx":
                return GRAPH_HELPERS.views.restExcel(
                    request, qset, fields,
                    name, format_size, localize=localize)
            elif format == "flat":
                return GRAPH_HELPERS.views.restFlat(
                    request, qset, fields,
                    name, format_size, localize=localize)
    
    @classmethod
    def on_rest_list_summary(cls, request, qset):
        cls._boundRest()
        summary_info = getattr(cls.RestMeta, "SUMMARY_FIELDS", {})
        output = objict()
        output.count = qset.count()
        for key, value in summary_info.items():
            if key == "sum":
                res = rh.getSum(qset, *value)
                if isinstance(res, dict):
                    output.update(res)
                else:
                    output[value[0]] = res
            elif key == "avg":
                for f in value:
                    output["avg_{}".format(f)] = rh.getAverage(qset, f)
            elif key == "max":
                for f in value:
                    output["max_{}".format(f)] = rh.getMax(qset, f)
            elif isinstance(value, dict):
                if "|" in key:
                    fields = key.split("|")
                    if len(fields) > 1:
                        lbl = fields[0]
                        action = fields[1]
                        field = None
                    if len(fields) > 2:
                        field = fields[2]
                else:
                    action = "count"
                    lbl = key
                    field = None
                act_qset = qset.filter(**value)
                if action == "count":
                    output[lbl] = act_qset.count()
                elif action == "sum":
                    output[lbl] = rh.getSum(act_qset, field)
                elif action == "avg":
                    output[lbl] = rh.getAverage(act_qset, field)
                elif action == "max":
                    output[lbl] = rh.getMax(act_qset, field)
        return GRAPH_HELPERS.restGet(request, output)

    @classmethod
    def getRestBatchCreateFilter(cls, item, exclude=["id", "pk", "created", "modified"], include=None):
        if isinstance(include, list):
            return {key: item[key] for key in include if key in item}
        # ignore related fields
        rfs = cls.get_related_name_fields()
        return {key: item[key] for key in item if key not in exclude and cls.hasField(key) and key not in rfs}

    @classmethod
    def on_rest_batch(cls, request, action):
        # this method is called when rest_batch='somme action'
        if not ALLOW_BATCHING or not getattr(cls.RestMeta, "CAN_BATCH", True):
            return GRAPH_HELPERS.restStatus(request, False, error="model does not allow batch actions")
        cls._boundRest()
        # if not request.member.hasPerm("can_batch_update"):
        #     raise re.PermissionDeniedException(f"batch updated not allowed by user")
        if action == "create":
            return cls.on_rest_batch_create(request)
        batch_ids = request.DATA.getlist("batch_ids", [])
        batch_id_field = request.DATA.get("batch_id_field", "pk")
        if batch_ids:
            q = {}
            q["{}__in".format(batch_id_field)] = batch_ids
            qset = cls.on_rest_list_query(request, cls.rw_objects().filter(**q))
        else:
            qset = cls.on_rest_list_query(request, cls.rw_objects().all())
        if action == "delete":
            can_delete = getattr(cls.RestMeta, "CAN_DELETE", False)
            if not can_delete:
                raise re.PermissionDeniedException(f"deletion not allowed for {cls.get_class_name()}", 438)
            count = qset.delete()[0]
            return GRAPH_HELPERS.restStatus(request, True, error="delete {} items".format(count))
        elif action == "update":
            update_fields = request.DATA.get(["batch_data", "batch_update"])
            if not isinstance(update_fields, dict):
                return GRAPH_HELPERS.restStatus(request, False, error="batch_update should be key/values")
            if {"id", "pk", "created", "password"} & update_fields.keys():
                return GRAPH_HELPERS.restStatus(request, False, error="field/s not allowed")
            has_meta = False
            for key in update_fields:
                if key.startswith("metadata"):
                    has_meta = True
                    break
            if has_meta or not request.member.is_superuser:
                count = 0
                for obj in qset:
                    obj.checkPermsAndSave(request, update_fields)
                    count += 1
            else:
                # only super users can do this
                count = qset.update(**update_fields)
            return GRAPH_HELPERS.restStatus(request, True, error="updated {} items".format(count))
        return GRAPH_HELPERS.restStatus(request, False, error="not implemented")

    @classmethod
    def on_rest_batch_create(cls, request):
        batch_data = request.DATA.get("batch_data")
        if isinstance(batch_data, str):
            batch_data = objict.fromJSON(batch_data)
        if isinstance(batch_data, dict):
            if "data" in batch_data:
                batch_data = batch_data["data"]
            if isinstance(batch_data, dict):
                batch_data = [batch_data]
        items = []
        for item in batch_data:
            obj = cls.createFromBatch(item)
            if obj:
                items.append(obj)
        return GRAPH_HELPERS.restList(request, items)

    @classmethod
    def createFromBatch(cls, item, request=None):
        obj = None
        item_filter = cls.getRestBatchCreateFilter(item)
        if not item_filter:
            raise Exception("requires item filter")
        rh.debug("createFromBatch", item_filter)
        obj = cls.ro_objects().filter(**item_filter).last()
        if obj is None:
            obj = cls()
            obj.checkPermsAndSave(request, item)
        return obj

    @classmethod
    def on_rest_create(cls, request, pk=None):
        # permissions are checked in the save routine
        can_create = getattr(cls.RestMeta, "CAN_CREATE", True)
        if not can_create:
            return GRAPH_HELPERS.restStatus(request, False, error="creation not allowed via rest for this model.")

        if hasattr(cls.RestMeta, "KEY_TO_FIELD_MAP"):
            kv = {}
            for k, v in list(cls.RestMeta.KEY_TO_FIELD_MAP.items()):
                if hasattr(request, k):
                    value = getattr(request, k)
                    if value is not None:
                        kv[v] = value
            obj = cls.createFromRequest(request, **kv)
        else:
            obj = cls.createFromRequest(request)
        graph = request.DATA.get("graph", "default")
        return obj.restGet(request, graph)

    @classmethod
    def _boundRest(cls):
        RestModel._setupGraphHelpers()

    @staticmethod
    def _setupGraphHelpers():
        if not GRAPH_HELPERS.views:
            views = importlib.import_module("rest.views")
            GRAPH_HELPERS.views = views
            GRAPH_HELPERS.restNotFound = views.restNotFound
            GRAPH_HELPERS.restStatus = views.restStatus
            GRAPH_HELPERS.restList = views.restList
            GRAPH_HELPERS.restGet = views.restGet
        if not GRAPH_HELPERS.get_request:
            mw = importlib.import_module("rest.middleware")
            GRAPH_HELPERS.get_request = mw.get_request

    @classmethod
    def get_rest_help(cls):
        output = objict()
        if cls.__doc__:
            output.doc = cls.__doc__.rstrip()
        else:
            output.doc = ""
        output.model_name = cls.get_class_name()
        output.fields = cls.rest_getQueryFields(True)
        output.graphs = {}
        if hasattr(cls, "RestMeta"):
            output.graph_names = list(getattr(cls.RestMeta, "GRAPHS", {}).keys())
            for key in output.graph_names:
                output.graphs[key] = cls.getGraph(key)
            output.no_show_fields = getattr(cls.RestMeta, "NO_SHOW_FIELDS", [])
            output.no_save_fields = getattr(cls.RestMeta, "NO_SAVE_FIELDS", [])
            output.search_fields = getattr(cls.RestMeta, "SEARCH_FIELDS", [])
        return output

    @classmethod
    def on_rest_request(cls, request, pk=None):
        # check if model id is in post
        request.rest_class = cls
        cls._boundRest()
        if pk is None and getattr(cls.RestMeta, "PK_FROM_FILTER", False):
            pk_fields = []
            key = cls.get_class_name().lower()
            key_p = f"{key}_id"
            pk_fields.append(key_p)
            # check if the cls has a field with the class name, (causes conflict)
            if not cls.get_field_type(key):
                pk_fields.append(key)
            pk = request.DATA.get(pk_fields, None, field_type=int)
        # generic rest request handler
        if pk is not None:
            using = getattr(cls.RestMeta, "RO_DATABASE", None)
            if using is None:
                using = getattr(cls.RestMeta, "DATABASE", None)
            if using:
                obj = cls.objects.using(using).filter(pk=pk).last()
            else:
                obj = cls.objects.filter(pk=pk).last()
            if obj is None:
                return GRAPH_HELPERS.views.restNotFound(request)
            if request.method == "GET":
                return obj.on_rest_get(request)
            elif request.method == "POST":
                return obj.on_rest_post(request)
            elif request.method == "DELETE":
                return obj.on_rest_delete(request)
            return GRAPH_HELPERS.views.restNotFound(request)

        if request.method == "GET":
            return cls.on_rest_list(request)
        elif request.method == "POST":
            if request.DATA.get("rest_batch"):
                return cls.on_rest_batch(request, request.DATA.get("rest_batch"))
            return cls.on_rest_create(request)
        return GRAPH_HELPERS.views.restNotFound(request)

    @classmethod
    def searchFromRequest(cls, request, qset):
        '''returns None if not foreignkey, otherswise the relevant model'''
        search_fields = getattr(cls.RestMeta, "SEARCH_FIELDS", None)
        search_terms = getattr(cls.RestMeta, "SEARCH_TERMS", None)
        search_join = getattr(cls.RestMeta, "SEARCH_JOIN", "and")
        q = request.DATA.get(["search", "q"])
        if q:
            qset = search.filter(qset, q, search_fields, search_terms, search_join)
        return qset

    @classmethod
    def rest_getWHITELISTED(cls):
        if hasattr(cls.RestMeta, "WHITELISTED"):
            return cls.RestMeta.WHITELISTED
        return cls.__RestMeta__.WHITELISTED

    @classmethod
    def rest_getQueryFields(cls, detailed=False):
        field_names = []
        all_fields = True
        if hasattr(cls.RestMeta, "QUERY_FIELDS"):
            field_names = cls.RestMeta.QUERY_FIELDS
            all_fields = "all_fields" in field_names

        if all_fields:
            for f in cls._meta.fields:
                if not f.name.endswith("_ptr") or f in cls.rest_getWHITELISTED():
                    field_names.append(f.name)
            if issubclass(cls, MetaDataModel):
                if detailed:
                    field_names.append("metadata")
                else:
                    field_names.append("properties__key")
                    field_names.append("properties__value")
        if detailed:
            output = []
            for f in field_names:
                if f == "metadata":
                    t = "MetaData"
                    fm = None
                else:
                    t = cls.get_field_type(f)
                    fm = cls.get_fk_model(f)
                info = {}
                info["name"] = f
                info["type"] = t
                if fm:
                    info["model"] = "{}.{}".format(fm._meta.app_label, fm.__name__)
                try:
                    fd = cls._meta.get_field(f)
                    if fd.choices:
                        info["choices"] = fd.choices
                    if fd.help_text:
                        info["help"] = fd.help_text()
                except Exception:
                    pass
                output.append(info)
            return output
        return field_names

    @classmethod
    def on_rest_set_query_defaults(cls, request):
        if not hasattr(request, "default_rest_filters"):
            request.default_rest_filters = None
            default_filters = getattr(cls.RestMeta, "LIST_DEFAULT_FILTERS", None)
            if default_filters:
                # make a copy so we can remove filters until the end
                request.default_rest_filters = objict.fromdict(default_filters)

    @classmethod
    def filterFromRequest(cls, request, qset):
        '''returns None if not foreignkey, otherswise the relevant model'''
        filter = request.DATA.pop("filter")
        # rh.debug("filterFromRequest")
        cls.on_rest_set_query_defaults(request)
        if not filter:
            return qset

        field_names = cls.rest_getQueryFields()
        q = {}

        """
        we can do customer filters but the name must be a allowed field
        and can only be one level deep ie no double "__" "group__member__owner"
        html select:
            name: "user_filter"
            field: "filter"
            options: [
                {
                    label: "Staff Only",
                    value: "is_staff:1"
                },
                {
                    label: "Online",
                    value: "is_online:1"
                },
                {
                    label: "Online",
                    value: "is_online:1"
                },
            ]
        """
        if not isinstance(filter, dict):
            filters = filter.split(';')
            filter = {}
            for f in filters:
                if ":" in f:
                    k, v = f.split(':')
                    if v in ["true", "True"]:
                        v = True
                    elif v in ["false", "False"]:
                        v = False
                    elif v == "null":
                        k = f"{k}__isnull"
                        v = True
                    filter[k] = v
        if filter:
            qset = cls.on_rest_handle_filter(qset, filter, request)
        now = datetime.now()
        for key in filter:
            if request.default_rest_filters:
                j = request.default_rest_filters.pop(key, None)
            name = key.split('__')[0]
            value = filter[key]
            if name in field_names and value != '__':
                if isinstance(value, str) and ':' in value and value.startswith('__'):
                    k, v = value.split(':')
                    key = key + k.strip()
                    value = v.strip()
                if key.endswith("__in") and ',' in value:
                    if value.startswith("["):
                        value = value[1:-1]
                    value = value.split(',')
                elif value in ["true", "True", "1"]:
                    value = 1
                elif value in ["false", "False", "0"]:
                    value = 0
                if isinstance(value, str) and "(" in value and ")" in value:
                    # this is a special function call
                    # rh.log_print(value)
                    if value.startswith("days("):
                        spos = value.find("(")+1
                        epos = value.find(")")
                        # rh.log_print(int(value[spos:epos]))
                        value = now + timedelta(days=int(value[spos:epos]))
                        # rh.log_print(now)
                        # rh.log_print(value)
                    elif value.startswith("hours("):
                        spos = value.find("(")+1
                        epos = value.find(")")
                        value = now + timedelta(hours=int(value[spos:epos]))
                    elif value.startswith("minutes("):
                        spos = value.find("(")+1
                        epos = value.find(")")
                        value = now + timedelta(minutes=int(value[spos:epos]))
                    elif value.startswith("seconds("):
                        spos = value.find("(")+1
                        epos = value.find(")")
                        value = now + timedelta(seconds=int(value[spos:epos]))
                    else:
                        continue
                if key.count('__') <= 4:
                    q[key] = value
            else:
                rh.log_print("filterFromRequest: invalid field: {} or {}".format(name, key))
        if q:
            # rh.debug("filtering...", q)
            qset = qset.filter(**q)
        return qset

    @classmethod
    def on_rest_handle_filter(cls, qset, filters, request=None):
        return qset

    @classmethod
    def filterRequestFields(cls, request, qset):
        """
        REQUEST_FIELDS which are fields attached to a request by middleware
        these fields can be used to auto filter requests
        """
        rfields = getattr(cls.RestMeta, "REQUEST_FIELDS", None)
        if rfields is not None:
            q = {}
            for fn in rfields:
                value = getattr(request, fn, None)
                if value is not None:
                    q[rfields[fn]] = value
            if bool(q):
                qset = qset.filter(**q)
        return qset

    ALLOWED_QUERY_OPERATORS = ["isnull", "gt", "gte", "lt", "lte", "contains", "icontains", "in", "startswith", "endswith"]

    @classmethod
    def queryFromRequest(cls, request, qset):
        """
        This will look at the query request and allow for basic filtering.

        QUERY_FIELDS_SPECIAL can be used to allow mappings of certain queries:
            ?group=iso
            {
                "group": "group__kind"
            }

        QUERY_FIELDS can be used to restrict what fields can be queried 
        and add special fields:

        ?group__kind=iso
        # this will allow "group__kind" but also support all other fields
        ["group__kind", "all_fields"]

        <field_name>=value
        or
        <field_name>__<operator>=value

        You cannot do <field_name>__<sub_field>__<operator> unless 
         <field_name>__<sub_field> is

        allowed <operator>s
            __gt
            __gte
            __lt
            __lte
            __icontains
            __in

        values can also be prefixed with special operators
            <field_name>=<operator>:value
        """
        q = {}
        exq = {}
        field_names = cls.rest_getQueryFields()
        query_keys = request.DATA.keys()
        special_keys = getattr(cls.RestMeta, "QUERY_FIELDS_SPECIAL", {})
        cls.on_rest_set_query_defaults(request)
        # rh.debug("queryFromRequest.key", query_keys, special_keys, field_names)
        for key in query_keys:
            # remove the key from defaults
            if request.default_rest_filters:
                j = request.default_rest_filters.pop(key, None)
                for drk in request.default_rest_filters:
                    if drk.startswith(f"{key}__"):
                        j = request.default_rest_filters.pop(drk, None)
                        break
            fn = key
            oper = None
            value = request.DATA.get(key)
            if value == "":
                continue
            if key in ["start", "size", "sort", "dr_start", "dr_end", "dr_eod", "dr_tz"]:
                continue
            if key.startswith("metadata__"):
                keys = key.split("__")
                meta_key = keys[1]
                meta_op = None
                meta_cat = None
                if len(keys) > 2:
                    if keys[2] in RestModel.ALLOWED_QUERY_OPERATORS:
                        meta_op = keys[2]
                    else:
                        meta_key = keys[2]
                        meta_cat = keys[1]
                if len(keys) > 3:
                    meta_op = keys[3]
                q["properties__key"] = meta_key
                if meta_cat:
                    q["properties__category"] = meta_cat
                if meta_op:
                    if RestModel.ALLOWED_QUERY_OPERATORS.index(meta_op) < 4:
                        q[f"properties__int_value__{meta_op}"] = value
                    else:
                        q[f"properties__value__{meta_op}"] = value
                else:
                    q["properties__value"] = value
                continue
            if key not in field_names:
                # check if special key
                if key in special_keys:
                    key = special_keys.get(key)
                    q[key] = value
                    continue
                # check if maybe this has a operator at end
                if "__" not in key:
                    continue
                fn = key[:key.rfind("__")]
                if fn not in field_names:
                    continue
                oper = key[key.rfind("__") + 2:]
                if oper not in RestModel.ALLOWED_QUERY_OPERATORS:
                    continue
                # this means we will allow it
                # fn = key[:key.rfind("__")]

            if value and isinstance(value, str) and value.startswith("__") and ":" in value:
                # this is another special way to filter via field=__
                oper, value = value[2:].split(':')
                if oper not in RestModel.ALLOWED_QUERY_OPERATORS:
                    continue
                key = "{}__{}".format(key, oper)
            if oper == "in":
                if isinstance(value, str) and ',' in value:
                    value = [a.strip() for a in value.split(',')]
            elif oper == "isnull":
                value = value in [True, "true", 1, "1"]
            elif oper is None and value in ["null", "None"]:
                key = "{}__isnull".format(key)
                value = True
            ft = cls.get_field_type(fn)
            if ft == "DateTimeField":
                value = rh.parseDateTime(value)
            elif ft == "IntegerField":
                value = rh.toInteger(value)
            q[key] = value
        if bool(q):
            exq = {key[:-4]: q[key] for key in q if key.endswith("__ne")}
            q = {key: q[key] for key in q if not key.endswith("__ne")}
            if bool(exq):
                qset = qset.exclude(**exq)
            if bool(q):
                qset = qset.filter(**q)
        return qset

    @classmethod
    def createFromRequest(cls, request, **kwargs):
        obj = cls()
        return obj.saveFromRequest(request, files=request.FILES, __is_new=True, **kwargs)

    @classmethod
    def createFromDict(cls, request=None, data=None, **kwargs):
        obj = cls()
        return obj.saveFromDict(request, data, __is_new=True, **kwargs)

    @classmethod
    def hasField(cls, fieldname):
        '''returns True if the fieldname exists in this model'''
        is_id = fieldname.endswith("_id")
        if is_id:
            fn = fieldname[:fieldname.rfind("_id")]
        for field in cls._meta.fields:
            if fieldname == field.name:
                return True
            if is_id and fn == field.name:
                return True
        return False

    @classmethod
    def get_class_name(cls, include_app=False):
        cname = cls.__name__
        if cls._meta.proxy:
            cname = cls.__bases__[0].__name__
        if include_app:
            return f"{cls._meta.app_label}.{cname}"
        return cname

    @classmethod
    def get_model_fields(cls):
        return cls._meta.fields

    @classmethod
    def get_model_field_names(cls):
        return [f.name for f in cls._meta.fields]

    @classmethod
    def has_model_field_name(cls, name):
        for f in cls._meta.fields:
            if f.name == name:
                return True
        return False

    @classmethod
    def get_field_type(cls, fieldname):
        '''returns the internal field type'''
        for field in cls._meta.fields:
            if fieldname == field.name:
                return field.get_internal_type()
        return None

    @classmethod
    def get_foreign_fields(cls):
        '''returns the internal field type'''
        return [field.name for field in cls._meta.fields if field.get_internal_type() == "ForeignKey"]

    @classmethod
    def get_related_name_fields(cls):
        return [f.related_name for f in cls._meta.related_objects]

    @classmethod
    def get_model_field_names(cls):
        return [f.name for f in cls._meta.get_fields()]

    @classmethod
    def get_fk_model(cls, fieldname):
        '''returns None if not foreignkey, otherswise the relevant model'''
        try:
            field = cls._meta.get_field(fieldname)
            return field.related_model
        except Exception:
            return None



# Generated by Django 4.0.2 on 2022-02-22 01:23

from django.conf import settings
import django.contrib.auth.models
import django.contrib.auth.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import rest.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('username', models.CharField(error_messages={'unique': 'A user with that username already exists.'}, help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.', max_length=128, unique=True, validators=[django.contrib.auth.validators.UnicodeUsernameValidator()], verbose_name='username')),
                ('first_name', models.CharField(blank=True, max_length=150, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=150, verbose_name='last name')),
                ('email', models.EmailField(blank=True, max_length=254, verbose_name='email address')),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.Group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.Permission', verbose_name='user permissions')),
            ],
            options={
                'db_table': 'auth_user',
            },
            bases=(models.Model, rest.models.RestModel),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='Group',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.CharField(blank=True, db_index=True, default=None, max_length=64, null=True)),
                ('name', models.CharField(db_index=True, max_length=200)),
                ('short_name', models.CharField(blank=True, default=None, max_length=60, null=True)),
                ('kind', models.CharField(db_index=True, default='org', max_length=80)),
                ('created', models.DateTimeField(auto_now=True)),
                ('is_active', models.BooleanField(blank=True, default=True)),
                ('children', models.ManyToManyField(related_name='parents', to='account.Group')),
                ('parent', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='groups', to='account.group')),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='Membership',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('role', models.CharField(blank=True, default='guest', max_length=64, null=True)),
                ('state', models.IntegerField(default=0)),
                ('group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='account.group')),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='NotificationRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('method', models.CharField(db_index=True, default='email', max_length=128)),
                ('from_addr', models.CharField(max_length=255)),
                ('subject', models.CharField(max_length=255)),
                ('body', models.TextField()),
                ('reason', models.TextField()),
                ('state', models.IntegerField(choices=[(0, 'new'), (-5, 'queued'), (1, 'sent'), (-10, 'failed')], db_index=True, default=0)),
                ('attempts', models.IntegerField(default=0)),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='Member',
            fields=[
                ('user_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to=settings.AUTH_USER_MODEL)),
                ('uuid', models.CharField(blank=True, db_index=True, default='', max_length=64)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('display_name', models.CharField(blank=True, default=None, max_length=64, null=True)),
                ('auth_code', models.CharField(blank=True, db_index=True, default=None, max_length=64, null=True)),
                ('password_changed', models.DateTimeField(blank=True, default=None, null=True)),
                ('last_activity', models.DateTimeField(blank=True, default=None, null=True)),
                ('requires_topt', models.BooleanField(blank=True, default=False, null=True)),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
                'abstract': False,
            },
            bases=('account.user', rest.models.RestModel, rest.models.MetaDataModel),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='NotificationAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('name', models.CharField(max_length=255)),
                ('mimetype', models.CharField(max_length=255)),
                ('data', models.TextField()),
                ('notification', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='account.notificationrecord')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='MembershipMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='account.membership')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='GroupMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='account.group')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='PasswordHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('password', models.CharField(max_length=255)),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='password_history', to='account.member')),
            ],
        ),
        migrations.CreateModel(
            name='NotificationMemberRecord',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('to_addr', models.CharField(db_index=True, max_length=255)),
                ('notification', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='to', to='account.notificationrecord')),
                ('member', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='notifications', to='account.member')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.AddField(
            model_name='membership',
            name='member',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='account.member'),
        ),
        migrations.CreateModel(
            name='MemberMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='account.member')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='MemberFeed',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('component', models.SlugField(blank=True, default=None, max_length=124, null=True)),
                ('component_id', models.IntegerField(blank=True, db_index=True, default=None, null=True)),
                ('kind', models.SlugField(max_length=32)),
                ('note', models.TextField(blank=True, default=None, null=True)),
                ('group', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.group')),
                ('member', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='feed', to='account.member')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='GroupFeed',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('component', models.SlugField(blank=True, default=None, max_length=124, null=True)),
                ('component_id', models.IntegerField(blank=True, db_index=True, default=None, null=True)),
                ('kind', models.SlugField(max_length=32)),
                ('note', models.TextField(blank=True, default=None, null=True)),
                ('group', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='feed', to='account.group')),
                ('member', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.member')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BounceHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('address', models.CharField(db_index=True, max_length=255)),
                ('kind', models.CharField(db_index=True, max_length=32)),
                ('reason', models.TextField(blank=True, default=None, null=True)),
                ('reporter', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('code', models.CharField(blank=True, default=None, max_length=32, null=True)),
                ('source', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('source_ip', models.CharField(blank=True, default=None, max_length=64, null=True)),
                ('user', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='bounces', to='account.member')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='AuthToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('token', models.TextField(db_index=True, unique=True)),
                ('role', models.CharField(blank=True, default=None, max_length=128, null=True)),
                ('member', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='auth_tokens', to='account.member')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
    ]




from django.apps import apps
from django.core.files.uploadedfile import *
from django.core.exceptions import ObjectDoesNotExist

import os
import tempfile
import string
import errno
from threading import Thread, Lock, Event

PresetParameters = {
    'width': {
        'description':	'Explicit width',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'height': {
        'description':	'Explicit height',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'maxwidth': {
        'description':	'Maximum width',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'maxheight': {
        'description':	'Maximum height',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'minwidth': {
        'description':	'Minimum width',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'minheight': {
        'description':	'Minimum height',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'maxaspect': {
        'description':	'Maximum aspect ratio (either direction)',
        'required':	False,
        'kind':		'F',
        'configurable':	True,
    },
    'aspect': {
        'description':	'Explicit aspect ratio',
        'required':	False,
        'kind':		'F',
        'configurable':	True,
    },
    'crop_x1': {
        'description':	'Crop left from top left',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'crop_y1': {
        'description':	'Crop top from top left',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'crop_x2': {
        'description':	'Crop right from top left',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
    'crop_y2': {
        'description':	'Crop bottom frop top left',
        'required':	False,
        'kind':		'I',
        'configurable':	True,
    },
}

def update_params(config, **kwargs):
    for p in kwargs:
        pn = p
        if pn[:2] == '__':
            pn = pn[2:]
        config['parameters'][pn]['value'] = kwargs[p]

class PipeThread(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.done = False
        self.exit = False
        try:
            self.delay_error = kwargs.pop('delay_error')
        except KeyError:
            self.delay_error = False
        try:
            self.daemon = kwargs.pop('daemon')
        except KeyError:
            self.daemon = False

        self.error = None
        self.tdata = ""
        self.tdata_lock = Lock()
        self.tdata_event = Event()

        self.write_t = Thread(target=self.writer)
        self.write_t.daemon = True
        self.write_t.start()

        if self.daemon:
            self.read_t = Thread(target=self.reader)
            self.read_t.daemon = True
            self.read_t.start()
        else:
            self.reader()
        if self.error:
            raise self.error

    def __del__(self):
        self.exit = True
        try:
            if self.read_t:
                self.read_t.join()
                self.read_t = None
        except:
            pass
        try:
            if self.write_t:
                self.write_t.join()
                self.write_t = None
        except:
            pass

    def wait(self, nohang=False):
        if self.daemon:
            if nohang:
                if self.read_t.is_alive():
                    return None

            self.read_t.join()
            if self.error:
                return -1
            return 0

    def kill(self):
        self.exit = True

    def writer(self):
        while True:
            self.tdata_event.wait()
            self.tdata_event.clear()
            if self.exit:
                return
            done = self.done
            with self.tdata_lock:
                data = self.tdata
                self.tdata = ""

            try:
                if data and not self.error:
                    self.write(data, *self.args, **self.kwargs)
            except Exception as e:
                self.error = e
                if not self.delay_error:
                    raise(e)
            if done or self.exit:
                return

    def reader(self):
        while True:
            data = self.read(*self.args, **self.kwargs)
            if self.exit or len(data) == 0:
                self.done = True
                if self.write_t:
                    self.tdata_event.set()
                    self.write_t.join()
                    self.write_t = None
                return

            with self.tdata_lock:
                self.tdata += data
            self.tdata_event.set()

class RenderProc:
    def __init__(self, pid, errfile=None):
        self.pid = pid
        self._errout = ""
        self._errlock = None
        if errfile:
            self.errread(errfile)

    def wait(self, nohang=False):
        opt = 0
        if nohang:
            opt |= os.WNOHANG
        while True:
            try:
                pid, sts = os.waitpid(self.pid, opt)
                if pid >= 0:
                    break
            except OSError as err:
                if err.errno == errno.EINTR:
                    if nohang:
                        return None
                    continue
                raise
        return os.WEXITSTATUS(sts)

    def kill(self):
        try:
            os.kill(self.pid, signal.SIGTERM)
        except OSError:
            pass

    @staticmethod
    def mkfifo():
        (fp, fname) = tempfile.mkstemp(prefix="t_")
        os.close(fp)
        os.unlink(fname)
        os.mkfifo(fname)
        return fname

    def _errworker(self, errfile):
        if type(errfile) == int:
            fd = errfile
            errfile = None
        else:
            fd = os.open(errfile, os.O_RDONLY)
        try:
            while True:
                data = os.read(fd, 10240)
                if len(data) == 0:
                    break
                with self._errlock:
                    self._errout += data
        finally:
            self._errlock = None
            try:
                os.close(fd)
            except:
                pass
            try:
                if errfile:
                    os.unlink(errfile)
            except:
                pass

    def errread(self, errfile):
        self._errlock = Lock()
        t = Thread(target=self._errworker, kwargs={'errfile': errfile})
        t.daemon = True
        t.start()

    def err(self):
        with self._errlock:
            return self._errout

    def errclear(self):
        with self._errlock:
            self._errout = ""

def get_rendition(item, use, name=None, fp=None):
    if item.kind == 'I' and use == 'still':
        use = 'original'
    if not hasattr(item, 'saved_renditions'):
        item.saved_renditions = {}
    if use in item.saved_renditions:
        rendition = item.saved_renditions[use]
    elif use == 'original':
        rendition = item.original()
        item.saved_renditions[use] = rendition
    else:
        rendition = item.renditions.all()
        if use:
            rendition = rendition.filter(use=use)
        if name:
            rendition = rendition.filter(name=name)
        rendition = rendition.order_by('-created').first()
        if rendition:
            item.saved_renditions[use] = rendition
    if rendition and not hasattr(rendition, 'filename'):
        if fp:
            if type(fp) in (str, str):
                rendition.filename = fp
                rendition.fp = open(fp, "r")
            elif isinstance(fp, UploadedFile):
                sfx = fp.name
                ok_chars = string.ascii_letters + string.digits + "_-."
                "".join(list(a in ok_chars and a or "_" for a in sfx))
                rendition.fp = tempfile.NamedTemporaryFile(suffix=sfx, prefix=(str(item.pk) + "_"))
                rendition.filename = rendition.fp.name
                for buf in fp.chunks():
                    rendition.fp.write(buf)
            else:
                rendition.fp = fp
                rendition.filename = rendition.fp.name
        else:
            rendition.fp = tempfile.NamedTemporaryFile(suffix=('.' + rendition.ext()), prefix=(str(item.pk) + "_"))
            try:
                rendition.get_file(rendition.fp)
                rendition.filename = rendition.fp.name
            except Exception:
                rendition.fp.close()
                rendition.fp = None
                rendition.filename = rendition.get_local_url()
        try:
            if rendition.fp:
                rendition.fp.flush()
        except IOError:
            pass
        if rendition.fp:
            rendition.fp.seek(0)
    return rendition

def new_rendition(item, renditiondef, fp, kind, width=None, height=None, bytes=None, meta={}, direct=False):
    doclose = False
    if direct:
        fname = fp
        fp = None
    elif type(fp) in (str, str):
        fp = open(fp, 'r')
        doclose = True
    elif fp:
        fp.flush()
        fp.seek(0)

    if not bytes:
        if fp:
            bytes = os.fstat(fp.fileno()).st_size
        else:
            bytes = 0

    model = apps.get_model('medialib', 'MediaItemRendition')
    rendition = model(mediaitem=item, name=renditiondef.name, use=renditiondef.use, rendition_definition=renditiondef, kind=kind, width=width, height=height, bytes=bytes)
    if fp:
        rendition.upload(fp)
    elif direct:
        rendition.url = fname + ".bob"
    rendition.save()

    if fp:
        rendition.fp = fp
        rendition.filename = fp.name
    if not hasattr(item, 'saved_renditions'):
        item.saved_renditions = {}
    if renditiondef.use in item.saved_renditions:
        item.saved_renditions[renditiondef.use + '_' + str(renditiondef.pk)] = item.saved_renditions[renditiondef.use]
    item.saved_renditions[renditiondef.use] = rendition

    for m in meta:
        rendition.set_meta(m, meta[m])

    return rendition

def close_files(item):
    for rendition in getattr(item, 'saved_renditions', {}):
        if hasattr(rendition, 'fp'):
            rendition.fp.close()
    item.saved_renditions = {}

def calc_size(params, origwidth, origheight, even=True, rotate=0):
    try:
        if rotate:
            params['rotate'] = rotate
            if rotate == 90 or rotate == 270:
                x = origheight;
                origheight = origwidth;
                origwidth = x;
        if params.get('width', 0)>0 and params.get('height', 0)>0:
            if params['width'] > origwidth:
                params.pop("width")
                params.pop("height")
                return params
            return params
        if params.get('width', 0)>0:
            if params['width'] > origwidth:
                params.pop("width")
                params.pop("height")
                return params
            params['height'] = origheight * params['width'] / origwidth
            return params
        if params.get('height', 0)>0:
            params['width'] = origwidth * params['height'] / origheight
            return params

        if params.get('maxwidth', 0) == 0 and params.get('maxheight', 0) == 0:
            return params
        if params.get('maxwidth', 0) == 0:
            params['maxwidth'] = 100000
        if params.get('maxheight', 0) == 0:
            params['maxheight'] = 100000
        if params.get('minwidth', 0) == 0:
            params['minwidth'] = 0
        if params.get('minheight', 0) == 0:
            params['minheight'] = 0

        if params.get('maxaspect', 0)>0 and (origwidth * params['maxheight']) / (origheight * params['maxwidth']) > params['maxaspect']:
            # too wide
            params['width'] = params['maxaspect'] * (origheight * params['maxwidth']) / params['maxheight']
            if params['width'] < params['minwidth']:
                params['width'] = params['minwidth']
            params['height'] = params['maxheight']
            left = (origwidth - params['width']) / params['maxaspect']
            params['crop'] = (left, 0, left+params['width']-1, origheight-1)
            return params
        if params.get('maxaspect', 0)>0 and (origheight * params['maxwidth']) / (origwidth * params['maxheight']) > params['maxaspect']:
            # too tall
            params['height'] = params['maxaspect'] * (origwidth * lib.thumb_height) / lib.thumb_width
            if params['height'] < params['minheight']:
                params['height'] = params['minheight']
            params['width'] = params['maxwidth']
            top = (origheight - params['height'] ) / params['maxaspect']
            params['crop'] = (0, top, origwidth-1, top+params['height']-1)
            return params

        if origwidth > origheight:
            # wider
            if origwidth > params['maxwidth']:
                params['width'] = params['maxwidth']
                params['height'] = origheight * params['maxwidth'] / origwidth
            else:
                params['width'] = origwidth
                params['height'] = origheight
            if params['width'] < params['minwidth']:
                params['width'] = params['minwidth']
            if params['height'] < params['minheight']:
                params['height'] = params['minheight']
            return params
        else:
            # taller
            if origheight > params['maxheight']:
                params['width'] = origwidth * params['maxheight'] / origheight
                params['height'] = params['maxheight']
            else:
                params['width'] = origwidth
                params['height'] = origheight
            if params['width'] < params['minwidth']:
                params['width'] = params['minwidth']
            if params['height'] < params['minheight']:
                params['height'] = params['minheight']
            return params
    finally:
        if 'width' in params and not params['width']:
            params.pop('width')
        if 'height' in params and not params['height']:
            params.pop('height')
        if even and 'width' in params and params['width'] % 2 == 1:
            params['width'] += 1
        if even and 'height' in params and params['height'] % 2 == 1:
            params['height'] += 1
        if 'aspect' in params:
            if params['height'] * params['aspect'] > params['width']:
                params['height'] = params['width'] / params['aspect']
            elif params['height'] * params['aspect'] < params['width']:
                params['width'] = params['height'] * params['aspect']
        if 'width' in params and 'height' in params and (not 'crop' in params) and origheight and origwidth and params['width'] * origheight != params['height'] * origwidth:
            if params['width'] * origheight > params['height'] * origwidth:
                # too tall
                want = int(params['height'] * origwidth / params['width'])
                chop1 = int((origheight - want)/2)
                # NOTE: always calculate bottom from original to prevent double rounding errors
                params['crop'] = (0, chop1, origwidth-1, origheight-1-chop1)
            else:
                # too wide
                want = int(params['width'] * origheight / params['height'])
                chop1 = int((origwidth - want)/2)
                # NOTE: always calculate left from original to prevent double rounding errors
                params['crop'] = (chop1, 0, origwidth-1-chop1, origheight-1)
        if 'height' in params:
            params['height'] = int(params['height'])
        if 'width' in params:
            params['width'] = int(params['width'])
        if 'crop_x1' in params or 'crop_x2' in params or 'crop_y1' in params or 'crop_y2' in params:
            params['crop'] = (
                params.get('crop_x1', 0),
                params.get('crop_y1', 0),
                origwidth - 1 - params.get('crop_x2', 0),
                origheight - 1 - params.get('crop_y2', 0),
            )
        return params

def save_meta(rendition, meta):
    model = apps.get_model('medialib', 'MediaMeta')
    for k in meta:
        obj, _ = model.objects.get_or_create(rendition=rendition, key=k)
        obj.value = str(meta[k])
        obj.save()





import sys
import re
import time
from datetime import date, datetime, timedelta
from .settings_helper import settings
import importlib
from django.db.models import Count, Q, Avg, Sum, Max, Min
from django.db.models.query import QuerySet
# from django.db.models.functions import TruncWeek, TruncMonth, TruncYear
from io import StringIO
import base64

from objict import objict
from .datem import *
from .log import getLogger
import getpass

from urllib.parse import urlparse


try:
    from fuzzywuzzy import fuzz
except:
    fuzz = None

# THIS NEEDS A NEW HOME
import shutil
import tempfile
import os
import random, string
import subprocess

ERROR_LOGGER = getLogger("errors", filename="error.log")
AUDIT_LOGGER = getLogger("auditlog", filename="auditlog.log")
DEBUG_LOGGER = getLogger("debug", filename="debug.log")
AUDITLOG_LOGGER_BY_USER = settings.get("AUDITLOG_LOGGER_BY_USER", False)
AUDITLOG_LOGGERS = settings.get("AUDITLOG_LOGGERS", {})
HELPER_CACHE = objict()
UNKNOWN = object()


def getLoggerByRequest(request):
    logger = AUDIT_LOGGER

    if AUDITLOG_LOGGER_BY_USER and request and request.user.is_authenticated:
        return getLogger(request.user.username, "{}.log".format(request.user.username))
    elif AUDITLOG_LOGGERS:
        for path, key in list(AUDITLOG_LOGGERS.items()):
            # print(("{}: {}".format(path, key)))
            if request.path.startswith(path):
                return getLogger(key.split('.')[0], filename=key)
    return logger


def getActiveRequest():
    if not HELPER_CACHE.get_request:
        mw = importlib.import_module("rest.middleware")
        HELPER_CACHE.get_request = mw.get_request
    return HELPER_CACHE.get_request()


def requestHasPerms(request, perms, group=None):
    # third party auth models
    if request.auth_model is not None:
        if request.auth_model.hasPerm(perms):
            return True, None, None

    if not request.user.is_authenticated:
        return False, "auth required", 401

    if request.member.hasPerm(perms):
        return True, None, None
    if group is None and hasattr(request, "group"):
        group = request.group
    if group:
        ms = request.member.getMembershipFor(group)
        if ms and ms.hasPerm(perms):
            return True, None, None
    return False, "permission denied", 402


def importModule(name):
    return importlib.import_module(name)


def log_print(*args):
    getLoggerByRequest(getActiveRequest()).info(*args)


def log(*args):
    getLoggerByRequest(getActiveRequest()).info(*args)


def debug(*args):
    DEBUG_LOGGER.info(*args)


def getLogRequestHeader(request, no_data=True):
    extra = []
    if bool(request):
        if hasattr(request, "ip"):
            extra.append(request.ip)
        extra.append(request.build_absolute_uri())
        if hasattr(request, "member") and request.member:
            extra.append("user: {}".format(request.member.username))
        if not no_data and hasattr(request, "DATA"):
            extra.append(request.DATA.asDict())
    return extra


def log_exception(*args):
    request = getActiveRequest()
    extra = getLogRequestHeader(request)
    ERROR_LOGGER.exception(*extra, *args)
    getLoggerByRequest(request).exception(*extra, *args)


def log_error(*args):
    request = getActiveRequest()
    extra = getLogRequestHeader(request)
    ERROR_LOGGER.error(*extra, *args)
    getLoggerByRequest(request).error(*extra, *args)


def log_stack(*args):
    request = getActiveRequest()
    import inspect
    extra = []
    for frame in inspect.stack():
        file, line, func = frame[1:4]
        extra.append(f"{file}, line {line} in function {func}")
    ERROR_LOGGER.error(*extra, *args)
    getLoggerByRequest(request).error(*extra, *args)


def setLogModel(self, component, pk):
    # this method allows for updating the component to associate a log with
    if not pk or (self._log_pk == pk):
        return
    self._log_component = component
    self._log_pk = pk
    if hasattr(self, "request_log"):
        self.request_log.component = component
        self.request_log.pkey = pk
        self.request_log.save()
        

def getHostname():
    try:
        return os.uname()[1]
    except:
        pass
    return "unknown"

def getUsername():
    return getpass.getuser()

class TemporaryDirectory(object):
    """Context manager for tempfile.mkdtemp() so it's usable with "with" statement."""
    def __enter__(self):
        self.name = tempfile.mkdtemp()
        return self.name

    def __exit__(self, exc_type, exc_value, traceback):
        shutil.rmtree(self.name)

class TemporaryFile(object):
    def __init__(self, suffix=".tmp"):
        temp_file = tempfile.NamedTemporaryFile(delete=False,   suffix=suffix)
        self.name = temp_file.name

    def delete(self):
        os.remove(self.name)

    def __enter__(self):
        return self.name

    def __exit__(self, type, value, traceback):
        self.delete()

# Some mobile browsers which look like desktop browsers.
RE_MOBILE = re.compile(r"(iphone|ipod|blackberry|android|palm|windows\s+ce)", re.I)
RE_DESKTOP = re.compile(r"(windows|linux|os\s+[x9]|solaris|bsd)", re.I)
RE_BOT = re.compile(r"(spider|crawl|slurp|bot)", re.I)
RE_SOCIAL = re.compile(r"(facebookexternalhit/[0-9]|LinkedInBot|Twitterbot|Pinterest|Google.*snippet)", re.I)
RE_SCRAPERS = re.compile(r"(FlipboardProxy|Slurp|PaperLiBot|TweetmemeBot|MetaURI|Embedly)", re.I)

RE_EMAIL = re.compile(r"[^@]+@[^@]+\.[^@]+", re.I)

DEBUG_DATETIME = False
if settings.DEBUG_DATETIME:
    DEBUG_DATETIME = settings.DEBUG_DATETIME


def graphBuilderInplace(part, field, graph):
    output = []
    if part not in graph:
        return output
    graph_part = graph[part]
    for f in graph_part:
        if type(f) is tuple:
            f1, f2 = f
            output.append(("{0}.{1}", "{2}").format(field, f1, f2))
        else:
            output.append("{0}.{1}".format(field, f))
    return output


def graphBuilder(root_graph, field, graph):
    for part in ["fields", "recurse_into"]:
        if part not in graph:
            continue
        graph_part = graph[part]
        if part not in root_graph:
            root_graph[part] = []
        root_part = root_graph[part]
        for f in graph_part:
            if type(f) is tuple:
                f1, f2 = f
                root_part.append(("{0}.{1}".format(field, f1), f2))
            else:
                root_part.append("{0}.{1}".format(field, f))
    return root_graph

def fuzzyMatch(a, b):
    if fuzz:
        if a and b:
            a = a.lower()
            b = b.lower()
            return max(fuzz.token_set_ratio(a,b),fuzz.partial_ratio(a,b))
        return 0
    print("MISSING FUZZWUZZY MODULE")
    return 100

def isValidEmail(email):
    return bool(RE_EMAIL.search(email))

def getProtocol(request):
    if request.is_secure():
        return "https://"
    return "http://"

def getSocialReferer(request):
    referer = getReferer(request)
    if referer and "://" in referer:
        r = referer.split("/")
        domain = r[2]
        if domain in ["t.co", "twitter.com"]:
            return "twitter"
        elif domain in ["facebook.com", "m.facebook.com"]:
            return "facebook"
        elif domain in ["linkedin.com", "lnkd.in"]:
            return "linkedin"
        elif domain in ["pinterest.com"]:
            return "pinterest"
        elif domain in ["plus.google.com", "plus.url.google.com"]:
            return "googleplus"
        elif domain in ["www.google.com"]:
            return "google"
        elif domain in ["bing.com", "www.bing.com"]:
            return "bing"
        return domain
    return referer

def getReferer(request):
    return request.META.get('HTTP_REFERER')

def getRemoteIP(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def parseHostName(url, root=False):
    if url.startswith("http"):
        parsed_url = urlparse(url)
        hname = parsed_url.netloc
    else:
        if "/" in url:
            url = url[:url.find('/')]
        hname = url
    if root and hname.count('.') > 1:
        return hname[hname.find(".")+1:]
    return hname


def getFromEmailForHost(host):
    email_domain = settings.EMAIL_DEFAULT_DOMAIN
    from_label = settings.EMAIL_FROM_LABEL
    if settings.EMAIL_SETTINGS:
        hostname = parseHostName(host, True)
        info = settings.EMAIL_SETTINGS.get(hostname, None)
        if info is not None:
            email_domain = info.get("email_domain", email_domain)
            from_label = info.get("from_label", from_label)
    return f"{from_label} <noreply@{email_domain}>"


def getUserAgent(request):
  # Some mobile browsers put the User-Agent in a HTTP-X header
  return request.META.get('HTTP_X_OPERAMINI_PHONE_UA') or \
         request.META.get('HTTP_X_SKYFIRE_PHONE') or \
         request.META.get('HTTP_USER_AGENT', '')

def getSocialAgent(request):
    user_agent = getUserAgent(request).lower()
    if "facebook" in user_agent:
        return "facebook"
    if "linkedin" in user_agent:
        return "linkedin"
    if "twitter" in user_agent:
        return "twitter"
    if "pinterest" in user_agent:
        return "pinterest"
    if "google" in user_agent:
        return "google"
    if "yahoo" in user_agent:
        return "yahoo"
    if "flipboard" in user_agent:
        return "flipboard"
    if "embedly" in user_agent:
        return "embedly"
    return "unknown"

def agentHas(request, keyword):
    return keyword.lower() in getUserAgent(request).lower()

def isSocialScraper(request):
    user_agent = getUserAgent(request)
    return bool(RE_SCRAPERS.search(user_agent))

def isSocialAgent(request):
    user_agent = getUserAgent(request)
    return bool(RE_SOCIAL.search(user_agent))

def isBotAgent(request):
    user_agent = getUserAgent(request)
    return bool(RE_BOT.search(user_agent))

def isMobile(request):
    user_agent = getUserAgent(request)
    return bool(RE_MOBILE.search(user_agent)) and not bool(RE_BOT.search(user_agent))

def isDesktopAgent(request):
    user_agent = getUserAgent(request)
    return not bool(RE_MOBILE.search(user_agent)) and bool(RE_DESKTOP.search(user_agent)) or bool(RE_BOT.search(user_agent))

def lowerKeys(x):
    if isinstance(x, list):
        return [lowerKeys(v) for v in x]
    elif isinstance(x, dict):
        return dict((k.lower(), lowerKeys(v)) for k, v in list(x.items()))
    return x

def mergeDicts(*args):
    context = {}
    for arg in args:
        context.update(arg)
    return context

def removeUnderscoreKeys(x):
    if isinstance(x, list):
        return [removeUnderscoreKeys(v) for v in x]
    elif isinstance(x, dict):
        return dict((k.replace('_', ''), removeUnderscoreKeys(v)) for k, v in list(x.items()))
    return x

def filterByDateRange(qset, request=None, start=None, kind=None, eod=None, field="created", end=None, zone=None):
    """
    "DateRangeStart": POSIX Start Date,
    "DateRangeEnd": POSIX End Date,
    "DateRangeEOD": Specify the UTC end of day for the range
    "DateRangeField": "created or modified",
    "DateRangeKind": "none, day, month, year",
    """
    if request:
        field = request.DATA.get(["daterangefield", "datefield"], field)
        # print(("field is: {}".format(field)))

        day = request.DATA.get("dateday", None)
        month = request.DATA.get("datemonth", None)
        year = request.DATA.get("dateyear", None)
        if day or month or year:
            # print(("running by day, month, year for field: {}".format(field)))
            qf = {}
            if day:
                qf["{}__day".format(field)] = day
            if month:
                qf["{}__month".format(field)] = month
            if year:
                qf["{}__year".format(field)] = year
            # print(qf)
            return qset.filter(**qf)

        start = request.DATA.get(["daterangestart", "date_start"], start)
        if not start:
            return qset
        start = parseDate(start)
        # print(("start is: {}".format(start)))
        zone = request.DATA.get("daterangezone", zone)
        eod = request.DATA.get("daterangeeod", eod, field_type=int)
        if eod == -1:
            zone = None
            eod = None
        elif eod is None and request.group:
            zone = request.group.timezone
            eod = request.group.getEOD(onday=start, in_local=True)
        elif eod is None:
            eod = 0
        end = request.DATA.get(["daterangeend", "date_end"], end)
        if end is None:
            end = start + timedelta(days=1)
        else:
            end = parseDate(end)

        if end <= start:
            end = start + timedelta(days=1)

        fields = request.DATA.get("daterangefields", None, field_type=list)
        if fields:
            start_field = fields[0]
            stop_field = fields[1]
            if start.hour == 0:
                # this is a hack on a straight date, lets move zone to 8 hours
                start = start - timedelta(hours=8)
            qf = {"{0}__lte".format(start_field):start, "{0}__gte".format(stop_field):start}
            return qset.filter(**qf)
        kind = request.DATA.get(["date_kind"], kind)
        # print "kind is: {}".format(kind)
        # zone = request.DATA.get("daterangezone", zone)
        # print "zone is: {}".format(zone)
        # print "eod is: {}".format(eod)
        # # print eod
        start, end = getDateRange(start=start, end=end, kind=kind, eod=eod, zone=zone)
        # print start
        # print end
    else:
        start, end = getDateRange(start=start, end=None, kind=kind, eod=eod, zone=zone)
        # print start
        # print end

    if not field:
        field = "created"

    qf = {"{0}__gte".format(field):start, "{0}__lt".format(field):end}
    return qset.filter(**qf)




def diffMinutes(t1, t2):
    diff = t1 - t2
    days, seconds = diff.days, diff.seconds
    hours = (days * 24)
    return (seconds * 60) + (hours / 60)

def diffHours(t1, t2):
    diff = t1 - t2
    days, seconds = diff.days, diff.seconds
    hours = (days * 24)
    minutes = seconds * 60
    return (minutes * 60) + hours

def getContext(request, *args, **kwargs):
    version = settings.VERSION
    if settings.DEBUG:
        version = "{0}.{1}".format(version, time.time())

    c = {
        "version":version,
        "SITE_LABEL":settings.SITE_LABEL,
        "SITE_LOGO":settings.SITE_LOGO,
        "SERVER_NAME":settings.SERVER_NAME,
        "BASE_URL":settings.BASE_URL,
        "DISCLAIMER": settings.REST_DISCLAIMER,
        "COMPANY_LABEL": settings.COMPANY_LABEL,
        "UNSUBSCRIBE_URL": settings.get("UNSUBSCRIBE_URL", f"{settings.BASE_URL}api/account/unsubscribe")
    }

    if request:
        c["protocol"] = getProtocol(request)
        c["request"] = request

    for k, v in kwargs.items():
        c[k] = v
    return c

def getAppNames():
    from django.apps import apps
    return [app_config.name for app_config in apps.get_app_configs()]

def getAllModels(app_name=None):
    from django.apps import apps
    if not app_name:
        return apps.get_models()
    return apps.get_app_config(app_name).get_models()


def find_nth(haystack, needle, n):
    start = haystack.find(needle)
    while start >= 0 and n > 1:
        start = haystack.find(needle, start+len(needle))
        n -= 1
    return start


def getSetting(key, default=None):
    return settings.get(key, default)


def filterByDates(qset, start=None, end=None, date_field="created"):
    q = {}
    if start:
        q["{}__gte".format(date_field)] = start

    if end:
        q["{}__lte".format(date_field)] = end
    if q:
        return qset.filter(**q)
    return qset


def getAverage(qset, field_name):
    res = qset.aggregate(avg_result=Avg(field_name))
    if "avg_result" in res and res["avg_result"] != None:
        return res["avg_result"]
    return 0.0


def getMin(qset, field_name):
    res = qset.aggregate(min_result=Min(field_name))
    if "min_result" in res and res["min_result"] != None:
        return res["min_result"]
    return 0.0


def getMax(qset, field_name):
    res = qset.aggregate(max_result=Max(field_name))
    if "max_result" in res and res["max_result"] != None:
        return res["max_result"]
    return 0.0


def getSum(qset, *args):
    params = {}
    for field_name in args:
        key = "sum_{}".format(field_name)
        params[key] = Sum(field_name)
    # print params
    res = qset.aggregate(**params)
    results = objict()
    for field_name in args:
        key = "sum_{}".format(field_name)
        value = res.get(key, 0)
        if value is None:
            value = 0
        results[field_name] = value
    if len(args) == 1:
        return list(results.values())[0]
    return results


def getGroupLeaders(Model, timeframe="week", date_field="created", 
                    size=10, field="id", action="count", out_field="value"):
    today = datetime.now()
    # FIXME: yes timerange to deal with timezone issues
    q = {f"{date_field}__year": today.year}
    if timeframe == "week":
        q[f"{date_field}__week"] = today.isocalendar()[1]
    elif timeframe == "month":
        q[f"{date_field}__month"] = today.month
    elif timeframe == "day":
        q[f"{date_field}__day"] = today.day
    qset = Model.objects.filter(**q).values('group__name', 'group')
    if field == "id":
        return list(qset.annotate(value=Count('id')).order_by('-value')[:size])
    field_key = f"{field}_{action}"
    if out_field:
        field_key = out_field
    aq = {}
    if action == "sum":
        aq[field_key] = Sum(field)
    elif action == "average":
        aq[field_key] = Avg(field)
    else:
        aq[field_key] = Count(field)
    return list(qset.annotate(**aq).order_by(f"-{field_key}")[:size])


def countOccurences(qset, field_name):
    output = objict()
    for item in list(qset.values(field_name).annotate(count=Count(field_name))):
        output[item[field_name]] = item["count"]
    return output


def filterOR(qset, *args):
    # filterOR(qset, dict(key1=val1), dict(key2=val2))
    q = Q(**args[0])
    for arg in args[1:]:
        q = q | Q(**arg)
    if not isinstance(qset, QuerySet):
        return qset.objects.filter(q)
    return qset.filter(q)

def updateModelFromDict(model, data):
    for key in data:
        setattr(model, key, data[key])
    return model

_KEY_NOTFOUND = object()

def getValueForKeys(data, key, default=None):
    # helper method for getting first key value from list
    if type(key) is list:
        for k in key:
            v = getValueForKeys(data, k, _KEY_NOTFOUND)
            if v != _KEY_NOTFOUND:
                return v
        return default

    if "." in key:
        keys = key.split('.')
        for k in keys:
            data = getValueForKeys(data, k, _KEY_NOTFOUND)
            if data is _KEY_NOTFOUND:
                return default
        return data

    if data is None:
        return None
    return data.get(key, default)

def dictToString(d, no_truncate=False):
    from io import StringIO
    output = StringIO()
    prettyWrite(d, output, no_truncate=no_truncate)
    out = output.getvalue()
    output.close()
    return out

def prettyPrint(d, f=sys.stdout, indent=4, banner=None):
    return prettyWrite(d, f, indent, banner)


try:
    import phonenumbers
except:
    phonenumbers = None

def isPhone(value):
    try:
        int(value.replace('-', '').replace('.', '').replace(' ', ''))
        return True
    except:
        pass
    return False

def normalizePhone(value):
    if phonenumbers:
        try:
            x = phonenumbers.parse(value, "US")
        except:
            print(value)
            x = phonenumbers.parse(value, None)
        return phonenumbers.format_number(x, phonenumbers.PhoneNumberFormat.INTERNATIONAL)
    return value.replace(' ', '').replace('.', '-').replace('(', '').replace(')', '-')


from decimal import Decimal
PRETTY_INDENT = 2
PRETTY_MAX_VALUE_LENGTH = 200
PRETTY_MAX_LINES = 400
PRETTY_MAX_LENGTH = 9000

def prettyWrite(d, f=None, indent=PRETTY_INDENT, banner=None, line_count=0, no_truncate=False):
    std_output = False
    if f is None:
        std_output = True
        f = StringIO()

    prev = None
    if banner:
        f.write('---- BEGIN {} ----\n'.format(banner))
    if type(d) is list:
        prev = False
        f.write('[')
        for i in d:
            if prev:
                line_count += 1
                f.write(',\n')
            else:
                line_count += 1
                f.write('\n')

            pos = 0
            if hasattr(f, "len"):
                pos = f.len

            if not no_truncate and (line_count > PRETTY_MAX_LINES or pos >= PRETTY_MAX_LENGTH):
                f.write('{}"...truncated"'.format(' ' * indent))
                break
            prev = True
            if type(i) is bool:
                i = int(i)
            if type(i) in [str, str]:
                if not no_truncate and len(i) >= PRETTY_MAX_VALUE_LENGTH:
                    f.write('{}"{}...truncated"'.format(' ' * indent, i[:PRETTY_MAX_VALUE_LENGTH-20]))
                else:
                    f.write('{}"{}"'.format(' ' * indent, i))
            elif type(i) is list or isinstance(i, dict):
                f.write(' ' * (indent))
                line_count = prettyWrite(i, f, indent+PRETTY_INDENT, line_count=line_count)
            elif type(i) is Decimal:
                f.write('{}{}'.format(' ' * indent, str(i)))
            else:
                f.write('{}"{}"'.format(' ' * indent, i))
        line_count += 1
        f.write('\n')
        f.write(' ' * (indent-PRETTY_INDENT))
        f.write(']')
    elif isinstance(d, dict):
        f.write('{')
        for key, value in list(d.items()):
            if prev:
                line_count += 1
                f.write(',\n')
            else:
                line_count += 1
                f.write('\n')

            pos = 0
            if hasattr(f, "len"):
                pos = f.len

            depth = indent / PRETTY_INDENT
            if not no_truncate and  (depth > 1 and (line_count > PRETTY_MAX_LINES or pos >= PRETTY_MAX_LENGTH)):
                f.write('{}"truncated":"...truncated"\n'.format(' ' * indent))
                break
            prev = True
            if type(key) in [str, str]:
                f.write('{}"{}":'.format(' ' * indent, key))
            else:
                f.write('{}{}: '.format(' ' * indent, str(key)))
            if type(value) is list or isinstance(value, dict):
                f.write(' ')
                line_count = prettyWrite(value, f, indent+PRETTY_INDENT, line_count=line_count)
            else:
                if type(value) in [str, str]:
                    if not no_truncate and len(value) >= PRETTY_MAX_VALUE_LENGTH:
                        f.write(' "{}...truncated"'.format(value[:PRETTY_MAX_VALUE_LENGTH-20]))
                    else:
                        f.write(' "{}"'.format(value))
                elif type(value) in [datetime, time]:
                    f.write(' "{}"'.format(value))
                elif type(value) is Decimal:
                    f.write(' {}'.format(str(value)))
                else:
                    f.write(' {}'.format(value))
        line_count += 1
        f.write('\n')
        f.write(' ' * (indent-PRETTY_INDENT))
        f.write('}')
    else:
        f.write(str(d))

    if banner:
        f.write('\n---- END {} ----\n'.format(banner))
    if indent == PRETTY_INDENT:
        f.write("\n")
    if std_output:
        sys.stdout.write(f.getvalue())
        f.close()
    return line_count

def randomKey(count=6):
    return ''.join(random.choice(string.uppercase + string.digits) for x in range(count))

def getStackString():
    import traceback
    return str(traceback.format_exc())

def isPidRunning(pid):
    if sys.platform.startswith("linux"):
        return os.path.exists("/proc/{}".format(pid))
    try:
        os.kill(pid, 0)  # or signal.SIGKILL
    except OSError as err:
        return False
    return True

def getPidFromFile(pid_file):
    pid = 0
    if os.path.exists(pid_file):
        with open(pid_file, "r") as f:
            pid = f.read()
            if pid.isdigit():
                pid = int(pid)
    return pid


def sudoCMD(cmd, as_user=None, test_first=False, return_output=False):
    if as_user and as_user == getUsername():
        sudo_cmd = []
        if isinstance(cmd, list):
            sudo_cmd.extend(cmd)
        else:
            sudo_cmd.append(cmd)
        return subprocess.Popen(sudo_cmd, close_fds=True)
    test_sudo_cmd = ["sudo", "-lU"]
    # -n for non-interactive, will print error if password required
    sudo_cmd = ["sudo", "-n"]
    if as_user:
        test_sudo_cmd.append("-lU")
        test_sudo_cmd.append(as_user)
        sudo_cmd.append("-u")
        sudo_cmd.append(as_user)
    else:
        test_sudo_cmd.append("root")
    if isinstance(cmd, list):
        test_sudo_cmd.extend(cmd)
        sudo_cmd.extend(cmd)
    else:
        test_sudo_cmd.append(cmd)
        sudo_cmd.append(cmd)
    if test_first:
        process = subprocess.Popen(test_sudo_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = process.communicate()
        if err.strip():
            raise Exception("WARNING: cannot run {}, we don't have sudo rights".format(test_sudo_cmd))
    if return_output:
        proc = subprocess.Popen(sudo_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return proc.communicate()
    return subprocess.Popen(sudo_cmd, close_fds=True)


def local_block_ip(ip, **kwargs):
    record_firewall_action(ip, "add", settings.hostname, kwargs.get("metadata", None))
    if settings.LOCAL:
        return True
    return sudoCMD([settings.BLOCK_IP_CMD, "add", ip])


def local_unblock_ip(ip, **kwargs):
    record_firewall_action(ip, "delete", settings.hostname, kwargs.get("metadata", None))
    if settings.LOCAL:
        return True
    return sudoCMD([settings.BLOCK_IP_CMD, "delete", ip])


def blockIP(ip, **kwargs):
    if settings.FIREWALL_GLOBAL_BLOCK:
        from taskqueue.models import Task
        Task.Publish("incident", "firewall_block", dict(ip=ip), channel="tq_broadcast")
        # record_firewall_action(ip, "add", "all", kwargs.get("metadata", None))
        return True
    return local_block_ip(ip, **kwargs)


def unblockIP(ip, **kwargs):
    # for global Task.Publish("incident", "global_unblock", {ip:"x.x.x.x"}, channel="tq_broadcast")
    if settings.FIREWALL_GLOBAL_BLOCK:
        from taskqueue.models import Task
        Task.Publish("incident", "firewall_unblock", dict(ip=ip), channel="tq_broadcast")
        # record_firewall_action(ip, "delete", "all", kwargs.get("metadata", None))
        return True
    return local_unblock_ip(ip, **kwargs)


def record_firewall_action(ip, action, hostname, metadata):
    from incident.models import Event
    title = f"FIREWALL: {ip} {action} on {hostname}"
    Event.createFromDict(None, dict(
        hostname=hostname,
        reporter_ip=ip,
        description=title,
        details=title,
        category="firewall",
        level=6,
        metadata=metadata
    ))


def getBlockedIPs():
    return sudoCMD([settings.BLOCK_IP_CMD, "list"])


def getBlockedHosts():
    blocked = []
    if not os.path.exists("/etc/hosts.deny"):
        return blocked
    with open("/etc/hosts.deny", 'r') as f:
        for line in f.readlines():
            if line.startswith("#"):
                continue
            if line.startswith("ALL:"):
                line = line.split("L:")[1]
            blocked.append(line.strip())
    return blocked


def toInteger(value):
    if isinstance(value, str):
        if "." in value:
            return int(float(value))
    if isinstance(value, list):
        return value
    return int(value)


def toString(value):
    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, bytearray):
        value = value.decode("utf-8")
    elif isinstance(value, (int, float)):
        value = str(value)
    return value

def toBytes(value):
    if isinstance(value, str):
        value = value.encode("utf-8")
    elif isinstance(value, bytearray):
        value = bytes(value)
    return value

def toByteArray(value):
    if isinstance(value, bytearray):
        return value
    elif isinstance(value, str):
        return bytearray(value, 'utf-8')
    elif isinstance(value, bytes):
        return bytearray(value)
    return value

def toHex(value):
    return toBytes(value).hex().upper()

def hexToByteArray(value):
    return bytearray.fromhex(value)

def hexToString(value):
    return bytes.fromhex(value).decode('utf-8')

def toBase64(value, as_str=True):
    if as_str:
        return toString(base64.b64encode(toBytes(value)))
    return base64.b64encode(toBytes(value))

def fromBase64(value):
    return toString(base64.b64decode(value))

def removeNonAscii(input_str):
    """Remove all non-ASCII characters from the input string."""
    return ''.join(char for char in input_str if 0 <= ord(char) < 128)


def dictDeepUpdate(dict1, dict2):
    for key, value in dict2.items():
        if isinstance(value, dict):
            if key not in dict1:
                dict1[key] = objict()
            dictDeepUpdate(dict1[key], value)
        else:
            dict1[key] = dict2[key]
    return dict1
    


# Generated by Django 4.1.4 on 2023-11-28 23:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0016_authsession_buid'),
    ]

    operations = [
        migrations.RenameField(
            model_name='member',
            old_name='requires_topt',
            new_name='requires_totp',
        ),
    ]



"""
This is where you can put handlers for running async background tasks

Task.Publish("myapp", "on_tq_test")
"""
from datetime import datetime, timedelta
from auditlog.models import PersistentLog
from sessionlog.models import SessionLog
from rest import settings

AUDITLOG_PRUNE_DAYS = settings.get("AUDITLOG_PRUNE_DAYS", 90)
REST_LOG_PRUNE_DAYS = settings.get("REST_LOG_PRUNE_DAYS", 90)
ACCOUNT_LOG_PRUNE_DAYS = settings.get("ACCOUNT_LOG_PRUNE_DAYS", 640)
ASYNC_LOG_PRUNE_DAYS = settings.get("ASYNC_LOG_PRUNE_DAYS", 14)
SESSION_PRUNE_DAYS = settings.get("SESSION_PRUNE_DAYS", 180)

ACCOUNT_LOG_COMPONENTS = settings.get("ACCOUNT_LOG_COMPONENTS", ["account.Member", "account.Group"])


def on_cleanup(task):
    # cleanup log files in the var directory
    count = cleanupAuditLogs()
    if count > 0:
        task.log(f"deleted {count} audit logs")

    count = cleanupAsyncLogs()
    if count > 0:
        task.log(f"deleted {count} async logs")

    count = cleanupRestLogs()
    if count > 0:
        task.log(f"deleted {count} rest logs")

    count = cleanupAccountLogs()
    if count > 0:
        task.log(f"deleted {count} account logs")

    count = cleanupSessionLogs()
    if count > 0:
        task.log(f"deleted {count} session logs")


def cleanupAuditLogs():
    if AUDITLOG_PRUNE_DAYS:
        before = datetime.now() - timedelta(days=AUDITLOG_PRUNE_DAYS)
        return PersistentLog.objects.filter(when__lte=before).exclude(component__in=ACCOUNT_LOG_COMPONENTS).delete()[0]
    return 0


def cleanupRestLogs():
    if REST_LOG_PRUNE_DAYS:
        before = datetime.now() - timedelta(days=REST_LOG_PRUNE_DAYS)
        return PersistentLog.objects.filter(when__lte=before, component__in=["rest", "action"]).exclude(component__in=ACCOUNT_LOG_COMPONENTS).delete()[0]
    return 0


def cleanupAccountLogs():
    if ACCOUNT_LOG_PRUNE_DAYS:
        before = datetime.now() - timedelta(days=ACCOUNT_LOG_PRUNE_DAYS)
        return PersistentLog.objects.filter(when__lte=before, component__in=ACCOUNT_LOG_COMPONENTS).delete()[0]
    return 0


def cleanupAsyncLogs():
    if ASYNC_LOG_PRUNE_DAYS:
        before = datetime.now() - timedelta(days=ASYNC_LOG_PRUNE_DAYS)
        return PersistentLog.objects.filter(when__lte=before, action="async").delete()[0]
    return 0


def cleanupSessionLogs():
    SessionLog.objects.filter()
    before = datetime.now() - timedelta(days=SESSION_PRUNE_DAYS)
    return SessionLog.objects.filter(created__lte=before).delete()[0]



from django.db import models
from datetime import datetime, timedelta
from rest import datem as date_util
from rest import models as rm
from rest import settings
from objict import objict

"""
This is useful for when creating daily metrics that have an end of day that is not midnight!
"""

DEFAULT_TIMEZONE = settings.get("METRICS_OFFSET_TIMEZONE", "US/Pacific")


def generate_uuid(slug, date_string, component=None, component_pk=None, group=None):
    uuid = slug
    if group is not None and component is not None:
        uuid = f"{slug}.g.{group.id}.{component}.{component_pk}:{date_string}"
    elif group is not None:
        uuid = f"{slug}.g.{group.id}:{date_string}"
    elif component is not None:
        uuid = f"{slug}.c.{component}.{component_pk}:{date_string}"
    else:
        uuid = f"{slug}:{date_string}"
    return uuid


def generate_date(date, tz=DEFAULT_TIMEZONE, eod_hour=0, eod_minute=0, granularity="daily"):
    if date is None:
        date = datetime.utcnow()
    pi = objict(tz=tz, hour=eod_hour, minute=eod_minute)
    end = date_util.convertToLocalTime(pi.tz, date)
    if end.hour >= pi.hour and end.minute >= pi.minute:
        end = end + timedelta(days=1)
    end = end.replace(hour=pi.hour, minute=pi.minute, second=0, microsecond=0)
    utc_end = date_util.convertToUTC(pi.tz, end)
    return utc_end, generate_date_string(utc_end, granularity)


def generate_date_string(when, granularity="daily"):
    date_string = when.strftime("%Y-%m-%d")
    return date_string


def metric(keys, slug, data, **kwargs):
    """
    keys: a list of ordered key names that will be used to keep data mappings in order
          key/value position 0 is the only "indexed" value
    slug: the unique slug to identify the data for the metrics
    data: a dict of key/value pairs that has values for all the keys
    component: this can be any string the represents unique data but is usually django model path
    componenet_pk: a unique key for the component
    group: optional group to map to
    tz: timezone
    eod: end of day
    """
    tz = kwargs.get("tz", DEFAULT_TIMEZONE)
    eod_hour = kwargs.get("eod_hour", 0)
    eod_minute = kwargs.get("eod_minute", 0)
    group = kwargs.get("group", None)
    component = kwargs.get("component", None)
    component_pk = kwargs.get("component_pk", None)
    when = kwargs.get("when", None)

    utc_end, date_string = generate_date(when, tz, eod_hour, eod_minute)
    uuid = generate_uuid(slug, date_string, component, component_pk, group)

    defaults = objict(slug=slug, end=utc_end)

    if group is not None:
        defaults.group_id = group.id
    if component is not None:
        defaults.component = component
        defaults.component_pk = component_pk
    mets, created = EODMetrics.objects.get_or_create(
        uuid=uuid, defaults=defaults)
    mets.updateMetrics(keys, data, update_keys=created)
    return mets


def metrics_component_group_all(keys, slug, data, **kwargs):
    """
    this method does the same as above but will generate all 3 metrics
     - global/all metrics
     - group metrics
     - component metrics
    """
    tz = kwargs.get("tz", DEFAULT_TIMEZONE)
    eod_hour = kwargs.get("eod_hour", 0)
    eod_minute = kwargs.get("eod_minute", 0)
    group = kwargs.get("group", None)
    component = kwargs.get("component", None)
    component_pk = kwargs.get("component_pk", None)
    metric(keys, slug, data, tz=tz, eod_hour=eod_hour, eod_minute=eod_minute)
    if group is not None:
        metric(
            keys, slug, data, group=group,
            tz=tz, eod_hour=eod_hour, eod_minute=eod_minute)
    if component is not None:
        metric(
            keys, slug, data, group=group,
            component=component, component_pk=component_pk,
            tz=tz, eod_hour=eod_hour, eod_minute=eod_minute)
    return True


def get_metric(slug, end=None, group=None, **kwargs):
    tz = kwargs.get("tz", DEFAULT_TIMEZONE)
    eod_hour = kwargs.get("eod_hour", 0)
    eod_minute = kwargs.get("eod_minute", 0)
    group = kwargs.get("group", None)
    component = kwargs.get("component", None)
    component_pk = kwargs.get("component_pk", None)

    utc_end, date_string = generate_date(end, tz, eod_hour, eod_minute)
    uuid = generate_uuid(slug, date_string, component, component_pk, group)
    
    obj = EODMetrics.objects.filter(uuid=uuid).last()
    return obj.getMetrics()


def get_metrics(slug, count=7, end=None, group=None, **kwargs):
    """
    returns data ready for Chart.js
    'periods': ['y:2012', 'y:2013', 'y:2014']
    'data': {
        key: value,
        etc...
    }
    """
    tz = kwargs.get("tz", DEFAULT_TIMEZONE)
    eod_hour = kwargs.get("eod_hour", 0)
    eod_minute = kwargs.get("eod_minute", 0)
    group = kwargs.get("group", None)
    component = kwargs.get("component", None)
    component_pk = kwargs.get("component_pk", None)

    utc_end, date_string = generate_date(end, tz, eod_hour, eod_minute)
    utc_now = utc_end - timedelta(days=count)
    periods = []
    missing = []
    data = None
    for i in range(0, count):
        utc_now = utc_now + timedelta(days=1)
        date_string = generate_date_string(utc_now)
        uuid = generate_uuid(slug, date_string, component, component_pk, group)
        periods.append(date_string)
        obj = EODMetrics.objects.filter(uuid=uuid).last()
        if obj is None:
            missing.append(0)
            continue
        obj_metrics = obj.getMetrics()
        if not data:
            data = dict()
            for key in obj_metrics.keys():
                data[key] = [*missing]
        for key, value in obj_metrics.items():
            data[key].append(value)
    return objict(periods=periods, data=data)


class EODMetrics(models.Model, rm.RestModel):
    class RestMeta:
        GRAPHS = {
            "default": {
                "recurse_into": ["generic__component"]
            }
        }

    created = models.DateTimeField(auto_now_add=True, editable=True)
    # timeframe of metrics is always a day and the ending day
    end = models.DateTimeField(db_index=True)
    # unique uuid of metric
    uuid = models.SlugField(max_length=250, unique=True)
    # the slug
    slug = models.CharField(max_length=200, db_index=True, null=True, default=None)
    # auto delete objects older then this
    expires = models.DateTimeField(db_index=True, null=True, default=None)
    # granularity: year, month, week, day, hour
    granularity = models.CharField(max_length=64, db_index=True)
    # our generic component
    component = models.CharField(max_length=255, db_index=True, null=True, default=None)
    # generic component primary key
    component_pk = models.IntegerField(null=True, default=None, db_index=True)
    # allow to group metrics by a group
    group_id = models.IntegerField(null=True, default=None, db_index=True)

    # now we create a set of k/v 
    k1 = models.SlugField(max_length=64, null=True, default=None)
    v1 = models.IntegerField(default=0, db_index=True)

    k2 = models.SlugField(max_length=64, null=True, default=None)
    v2 = models.IntegerField(default=0, db_index=True)

    k3 = models.SlugField(max_length=64, null=True, default=None)
    v3 = models.IntegerField(default=0)

    k4 = models.SlugField(max_length=64, null=True, default=None)
    v4 = models.IntegerField(default=0)

    k5 = models.SlugField(max_length=64, null=True, default=None)
    v5 = models.IntegerField(default=0)

    k6 = models.SlugField(max_length=64, null=True, default=None)
    v6 = models.IntegerField(default=0)

    k7 = models.SlugField(max_length=64, null=True, default=None)
    v7 = models.IntegerField(default=0)

    k8 = models.SlugField(max_length=64, null=True, default=None)
    v8 = models.IntegerField(default=0)

    k9 = models.SlugField(max_length=64, null=True, default=None)
    v9 = models.IntegerField(default=0)

    k10 = models.SlugField(max_length=64, null=True, default=None)
    v10 = models.IntegerField(default=0)

    k11 = models.SlugField(max_length=64, null=True, default=None)
    v11 = models.IntegerField(default=0)

    k12 = models.SlugField(max_length=64, null=True, default=None)
    v12 = models.IntegerField(default=0)

    k13 = models.SlugField(max_length=64, null=True, default=None)
    v13 = models.IntegerField(default=0)

    k14 = models.SlugField(max_length=64, null=True, default=None)
    v14 = models.IntegerField(default=0)

    def getMetrics(self):
        metrics = objict()
        for i in range(1, 15):
            key = getattr(self, f"k{i}", None)
            if key is None:
                return metrics
            metrics[key] = getattr(self, f"v{i}", 0)
        return metrics

    def updateMetrics(self, keys, data, update_keys=True):
        params = {}
        index = 0
        for key in keys:
            if isinstance(data, list):
                v = data[index]
            else:
                v = data[key]
            index += 1
            vkey = f"v{index}"
            params[vkey] = models.F(vkey) + v
            if update_keys or True:
                params[f"k{index}"] = key
        EODMetrics.objects.filter(pk=self.pk).update(**params)



from rest import settings
if settings.LEGACY_SERIALIZER:
    from .serializers.legacy import *  # noqa: F401, F403
else:
    from .serializers.response import *  # noqa: F401, F403
from . import url_docs
from django.shortcuts import render
REST_PREFIX = settings.get("REST_PREFIX", "api/")


def showDocs(request):
    from . import urls
    apis, graphs = url_docs.getRestApis(urls.urlpatterns)
    for api in apis:
        api["url"] = "{REST_PREFIX}{0}".format(api["url"])
    return render(request, "rest_docs.html", {"apis": apis, "graphs":graphs})


def __autoResponse(data, **kwargs):
    return '{"status": false, "error": "system offline"}'


backUpResult = restResult
_returnResults = restResult


def __offline(request):
    if settings.OFFLINE_KEY and request.get("key_check") == settings.OFFLINE_KEY:
        # remove failed files
        if request.DATA.get("offline", field_type=bool, default=1):
            restResult = __autoResponse
        else:
            restResult = backUpResult


def csrf_failure(request, reason=""):
    return restStatus(request, False, error=f"csrf failure: {reason}")



# Generated by Django 4.1.4 on 2024-02-20 21:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('location', '0002_geoip_subnet_alter_geoip_ip'),
    ]

    operations = [
        migrations.AddField(
            model_name='address',
            name='postalcode_suffix',
            field=models.CharField(blank=True, default=None, max_length=32, null=True),
        ),
    ]



from ..exceptions import *
from .. import schedule as sched

import os
import tempfile
import subprocess
import shutil
import traceback
from medialib import utils
from rest.log import getLogger
logger = getLogger("medialib", filename="medialib.log")


ffmpeg_cmd = ['ffmpeg', '-loglevel', 'warning', '-y', ]
ffmpeg_threads = ['-threads', '4', ]

scheduled = []
cleanup_run = []


def run_cmd(args, stdin=None):
    fp = tempfile.TemporaryFile()
    if not stdin:
        stdin = os.open("/dev/null", os.O_RDONLY)
    proc = subprocess.Popen(args, close_fds=True, stderr=fp.fileno(), stdout=fp.fileno(), stdin=stdin)
    os.close(stdin)
    return (proc, fp)

def check_cmd(proc, fp, cmd=None):
    retval = proc.wait()
    fp.flush()
    fp.seek(0)
    bufl = fp.readlines()
    fp.close()
    buf = ""
    for l in bufl:
        l = utils.toString(l)
        if (l.startswith("FFmpeg version ") or
            l.startswith("  built on ") or
            l.startswith("  configuration: ") or
            l.startswith("  lib")):
            pass
        else:
            buf += l
    if retval:
        raise CmdError(retval, buf, cmd=cmd)
    return retval

def _params_to_args(params={}, extra_args=[], infile=None):
    args = []
    
    if 'height' in params and 'width' in params and 'size' not in params:
        if (params['width'] % 2) != 0:
            params['width'] = params['width'] - 1
        if (params['height'] % 2) != 0:
            params['height'] = params['height'] - 1
        params['size'] = '%dx%d' % (params['width'], params['height'])
    copy_params_pre = (
            ('seek_time', '-ss'),
        )
    copy_params = (
            ('video_codec', '-vcodec'),
            ('video_profile', '-vpre'),
            ('video_bitrate', '-b:v', '', 'k'),
            ('audio_codec', '-acodec'),
            ('audio_bitrate', '-b:a', '', 'k'),
            ('audio_channels', '-ac'),
            ('audio_freq', '-ar'),
            ('size', '-s'),
            ('frame', '-vframes'),
            ('frame_rate', '-r'),
            ('key_frame', '-g'),
            ('duration', '-t'),
            ('subme', '-subq'),
            ('merange', '-me_range'),
            ('video_container', '-f'),
            ('pix_fmt', '-pix_fmt'),
        )
    # if not params.get('no_threads', False):
    #     args += ffmpeg_threads
    pass_thru = int(params.get("pass_thru", 0))
    if pass_thru == 1:
        logger.info("pass_thru ignoring params", params)
        # only allow height to pass through
        new_params = {}
        for key in ["size", "video_codec"]:
            if key in params:
                new_params[key] = params[key]
        params = new_params

    for p in copy_params_pre:
        if p[0] in params:
            if type(params[p[0]]) == float:
                s = "%.5f" % params[p[0]]
            else:
                s = str(params[p[0]])
            if len(p) > 2:
                s = p[2] + s
            if len(p) > 3:
                s = s + p[3]
            args += [p[1], s]
    if infile:
        args += ['-i', infile]

    rotate = int(params.get('rotate', '0'))
    if rotate == 90:
        args += ['-vf', 'transpose=1']
    elif rotate == 180:
        args += ['-vf', 'vflip,hflip']
    elif rotate == 270:
        args += ['-vf', 'transpose=2']

    for p in copy_params:
        if p[0] in params:
            if type(params[p[0]]) == float:
                s = "%.5f" % params[p[0]]
            else:
                s = str(params[p[0]])
            if len(p) > 2:
                s = p[2] + s
            if len(p) > 3:
                s = s + p[3]
            args += [p[1], s]

    args += extra_args
    return args

def run(outfile, infile, params={}, extra_args=[], final_args=[]):
    cmd = ffmpeg_cmd + []
    cmd += _params_to_args(params, extra_args, infile)
    npass = params.get('pass', 1)
    currpass = 1
    tmpdir = tempfile.mkdtemp()
    prevdir = os.getcwd()
    os.chdir(tmpdir)
    try:
        while currpass <= npass:
            if npass == 1:
                cmdx = cmd + final_args + [outfile]
            elif currpass < npass:
                cmdx = cmd + ['-pass', str(currpass), '-f', 'rawvideo', '/dev/null']
            else:
                cmdx = cmd + ['-pass', str(currpass), ] + final_args + [outfile]
            logger.info("ffmpeg command", cmdx)
            (proc, fd) = run_cmd(cmdx)
            check_cmd(proc, fd, cmd=cmdx)
            currpass += 1
    except Exception as e:
        try:
            os.unlink(outfile)
        except:
            pass
        os.chdir(prevdir)
        raise e
    finally:
        os.chdir(prevdir)
        shutil.rmtree(tmpdir)

    return outfile

def schedule(outfile, infile, params={}, extra_args=[], final_args=[], cleanup=None, data=None):
    global scheduled
    scheduled.append({
        'params': params,
        'input': infile,
        'outfile': outfile,
        'extra_args': extra_args,
        'final_args': final_args,
        'cleanup': cleanup,
        'data': data,
    })
    sched.register_run(schedule_run)
    sched.register_cleanup(schedule_cleanup)

def schedule_run():
    global scheduled
    global cleanup_run
    while scheduled:
        this_run = [scheduled.pop(0)]
        i = 0
        while i < len(scheduled):
            if scheduled[i]['input'] == this_run[0]['input']:
                this_run.append(scheduled.pop(i))
            else:
                i += 1
        infile = this_run[0]['input']
        no_threads = False
        for one in this_run:
            if one['params'].get('no_threads', False):
                no_threads = True
        one['no_threads'] = no_threads
        for one in this_run:
            cmd = ffmpeg_cmd + _params_to_args(one['params'], one['extra_args'], infile) + one['final_args'] + [ one['outfile'] ]

        (proc, fd) = run_cmd(cmd)
        cleanup_run.append((proc, fd, this_run, ))

def schedule_cleanup():
    global cleanup_run
    ex = []
    for (proc, fd, this_run) in cleanup_run:
        try:
            check_cmd(proc, fd)
        except Exception as e:
            tb = getattr(e, 'stack', None)
            if not tb:
                tb = "".join(traceback.format_stack(None, 6)[:-1])
            ex.append((__name__, e, tb))
        while this_run:
            one = this_run.pop()
            if one['cleanup']:
                try:
                    one['cleanup'](one['data'], bool(ex))
                except Exception as e:
                    tb = getattr(e, 'stack', None)
                    if not tb:
                        tb = "".join(traceback.format_stack(None, 6)[:-1])
                    ex.append((one['cleanup'].__module__, e, tb))

    cleanup_run = []

    if ex:
        msg = "\n".join(('%s: %s\n\tSTACK\t: %s' % (x[0], str(x[1]), x[2].strip().replace("\n", "\n\t\t: ")) for x in ex))
        raise RenderError(msg)




from django.core.management.base import BaseCommand
import urls

from rest import url_docs

class Command(BaseCommand):
    help = "Inspect urlpatterns and their views looking for mismatched arguments"

    def handle(self, *labels, **options):
        apis = url_docs.getRestApis(urls.urlpatterns)

        for api in apis:
            print((api["url"]))
            print((api["doc"]))



from django.db import models
from django.db.transaction import atomic
from rest import models as rm
from rest import settings
from datetime import datetime, timedelta
import time
from .. import geolocate

GEOIP_LOOKUP_BY_SUBNET = settings.get("GEOIP_LOOKUP_BY_SUBNET", True)


class GeoIP(models.Model, rm.RestModel):
    class RestMeta:
        SEARCH_FIELDS = ["ip", "isp"]
        CAN_DELETE = False

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    hostname = models.CharField(max_length=255, blank=True, null=True, default=None)
    ip = models.CharField(max_length=64, db_index=True)
    subnet = models.CharField(max_length=64, db_index=True, default=None, null=True)
    isp = models.CharField(max_length=84)

    city = models.CharField(max_length=64, blank=True, null=True)
    state = models.CharField(max_length=64, blank=True, null=True)
    country = models.CharField(max_length=64, blank=True, null=True)
    postal = models.CharField(max_length=32, blank=True, null=True)

    lat = models.FloatField(default=0.0, blank=True)
    lng = models.FloatField(default=0.0, blank=True)

    def __str__(self):
        return f"<GeoIP: {self.ip} {self.lat}:{self.lng}"

    @property
    def age(self):
        return (datetime.now() - self.modified).total_seconds()
    
    def isStale(self, days=300):
        stale = self.modified + timedelta(days=days)
        return stale < datetime.now()

    def refreshFromDB(self):
        if self.pk and self.country is None and self.age < 30.0:
            time.sleep(0.5)
            self.refresh_from_db()

    def refresh(self, using=None):
        if self.ip == "127.0.0.1":
            self.isp = "local"
            self.hostname = "localhost"
            self.subnet = "127.0.0"
            self.modified = datetime.now()
            self.save(using=using)
            return
        # save now to stop double lookups
        if not self.pk:
            self.isp = "unknown"
            self.saveNow(using)
        res = geolocate.locateByIP(self.ip)
        if res is None:
            self.modified = datetime.now()
            self.save(using=using)
            return
        self.hostname = res.hostname
        if res.isp is None:
            self.isp = "unknown"
        else:
            self.isp = res.isp[:83]
        self.city = res.city
        self.state = res.state
        self.country = res.country
        self.postal = res.postal
        try:
            self.lat = float(res.latitude)
            self.lng = float(res.longitude)
        except Exception:
            pass
        self.save(using=using)

    @atomic
    def saveNow(self, using=None):
        # force saves now
        self.save(using=using)

    @classmethod
    def get(cls, ip, force_refresh=False, stale_after=300):
        return cls.lookup(ip, force_refresh, stale_after)

    @classmethod
    def lookup(cls, ip, force_refresh=False, stale_after=300, using=None):
        if isinstance(ip, str):
            ip = ip.strip()
        if not geolocate.isIP(ip):
            ip = geolocate.dnsToIP(ip)
            if ip is None:
                return None
        subnet = ip[:ip.rfind(".")]
        gip = GeoIP.objects.filter(ip=ip).first()
        if gip is None:
            gip = GeoIP(ip=ip, subnet=subnet)
            if GEOIP_LOOKUP_BY_SUBNET:
                subgip = GeoIP.objects.filter(subnet=subnet).last()
                if subgip:
                    gip.lat = subgip.lat
                    gip.lng = subgip.lng
                    gip.city = subgip.city
                    gip.state = subgip.state
                    gip.country = subgip.country
                    gip.postal = subgip.postal
                    gip.isp = subgip.isp[:83]
                    gip.save()
                else:
                    gip.refresh(using=using)
            else:
                gip.refresh(using=using)
        else:
            if force_refresh or gip.isStale(stale_after):
                gip.refresh(using=using)
            else:
                gip.refreshFromDB()
            if gip.subnet is None:
                gip.subnet = subnet
                gip.save()
        return gip

    @staticmethod
    def removeDuplicates():
        # Find all ip_addresses that have duplicates
        # would be good to figure out how to not create duplicates
        duplicates = GeoIP.objects.values('ip')\
                                  .annotate(ip_count=models.Count('id'))\
                                  .filter(ip_count__gt=1)
        for entry in duplicates:
            # Get the first instance of MyModel for this ip_address
            first_instance = GeoIP.objects.filter(ip=entry['ip'])\
                                            .order_by('id').first()
            # Find all related models
            for rel in GeoIP._meta.related_objects:
                if isinstance(rel, models.ForeignKey):
                    related_model = rel.related_model
                    fk_field = rel.field.name
                    # Update the ForeignKey in each related model
                    related_model.objects.filter(**{f"{fk_field}__ip": entry['ip']})\
                                         .exclude(**{fk_field: first_instance})\
                                         .update(**{fk_field: first_instance})
                elif isinstance(rel, models.ManyToOneRel):
                    related_model = rel.related_model
                    fk_field = rel.field.name
                    # print(f"{rel}.{related_model}.{fk_field}")
                    related_model.objects.filter(**{f"{fk_field}__ip": entry['ip']})\
                                         .exclude(**{fk_field: first_instance})\
                                         .update(**{fk_field: first_instance})

            # Delete the duplicate MyModel instances, keeping the first one
            GeoIP.objects.filter(ip=entry['ip'])\
                         .exclude(id=first_instance.id)\
                         .delete()
